from __future__ import annotations

import importlib.metadata
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING

from tbc_video_export.common.utils.metadata import get_url_from_metadata

if TYPE_CHECKING:
    from typing import Final

_metadata = importlib.metadata.metadata("tbc-video-export")

# substituted by poetry-dynamic-versioning when doing pyinstaller builds
__version__ = "0.1.2"

APPLICATION_NAME: Final = _metadata["Name"]
PROJECT_VERSION: Final = _metadata["Version"] if __version__ == "0.0.0" else __version__

PROJECT_CREDITS: Final = (
    "Credits:\n"
    "  Jitterbug\tDevelopment (https://github.com/JuniorIsAJitterbug)\n"
    "  Harry Munday\tProject Management (https://github.com/harrypm)\n"
)

PROJECT_SUMMARY: Final = f"{_metadata['Summary']}\n\n{PROJECT_CREDITS}"
PROJECT_URL: Final = _metadata["Home-Page"]
PROJECT_URL_ISSUES: Final = get_url_from_metadata("Issues")
PROJECT_URL_WIKI: Final = get_url_from_metadata("Wiki")
PROJECT_URL_DISCORD: Final = get_url_from_metadata("Discord")

PROJECT_URL_WIKI_COMMANDLIST: Final = f"{PROJECT_URL_WIKI}/Command-List"
PROJECT_URL_WIKI_PROFILES: Final = f"{PROJECT_URL_WIKI}/FFmpeg-Profiles"


CURRENT_TIMESTAMP: Final = datetime.now().strftime("%y-%m-%d_%H%M%S%f")[:-3]
EXPORT_CONFIG_FILE_NAME: Final = Path(f"{APPLICATION_NAME}.json")

PIPE_BUFFER_SIZE: Final = 4 * 1024 * 1024  # 4MB

MINIMUM_TERMINAL_HEIGHT: Final = 30
MINIMUM_TERMINAL_WIDTH: Final = 120

TWO_STEP_OUT_FILE_LUMA_SUFFIX: Final = "luma"

FFMPEG_VIDEO_MAP: Final = "[v_out]"

# Ubuntu 22.04 uses FFmpeg 4.4.1 which does not support the new format
FFMPEG_USE_OLD_MERGEPLANES: Final = True


# Windows 10 conhost uses unicode-less fonts by default :'(
# SUCCESS_SYMBOL = "✓"
# ERROR_SYMBOL = "🗙"
# RUNNING_SYMBOLS: tuple[str, ...] = ("⠋", "⠙", "⠹", "⠸", "⠼", "⠴", "⠦", "⠧", "⠇", "⠏")

SUCCESS_SYMBOL: Final = "●"
ERROR_SYMBOL: Final = "○"
RUNNING_SYMBOLS: Final[tuple[str, ...]] = ("/", "-", "\\", "|")

# for NT ANSI enabling
NT_STD_OUTPUT_HANDLE: Final = -11
NT_ENABLE_PROCESSED_OUTPUT: Final = 0x1
NT_ENABLE_WRAP_AT_EOL_OUTPUT: Final = 0x2
NT_ENABLE_VIRTUAL_TERMINAL_PROCESSING: Final = 0x4

# for NT named pipes
NT_NAMED_PIPE_MAX_INSTANCES: Final = 1
NT_NAMED_PIPE_TIMEOUT: Final = 0
NT_NAMED_PIPE_BUFFER_SIZE: Final = 65536

# for NT proc snapshots
NT_TH32CS_SNAPPROCESS: Final = 0x2



from __future__ import annotations

import asyncio
import os
from functools import cached_property
from typing import TYPE_CHECKING

from tbc_video_export.common.enums import PipeType, ProcessName
from tbc_video_export.common.utils import FlatList
from tbc_video_export.process.wrapper.wrapper import Wrapper

if TYPE_CHECKING:
    from tbc_video_export.process.wrapper.pipe import Pipe
    from tbc_video_export.process.wrapper.wrapper_config import WrapperConfig
    from tbc_video_export.program_state import ProgramState


class WrapperLDDropoutCorrect(Wrapper):
    """Wrapper for ld-dropout-correct."""

    def __init__(self, state: ProgramState, config: WrapperConfig[None, Pipe]) -> None:
        super().__init__(state, config)
        self._config = config

    def post_fn(self) -> None:  # noqa: D102
        pass

    @property
    def command(self) -> FlatList:  # noqa: D102
        return FlatList(
            (
                self.binary,
                "-i",
                self._state.file_helper.tbcs[self._config.tbc_type],
                "--input-json",
                self._state.file_helper.tbc_json.file_name,
                "--output-json",
                os.devnull,
                self._config.output_pipes.out_path,
            )
        )

    @cached_property
    def process_name(self) -> ProcessName:  # noqa: D102
        return ProcessName.LD_DROPOUT_CORRECT

    @cached_property
    def supported_pipe_types(self) -> PipeType:  # noqa: D102
        return PipeType.NULL | PipeType.OS

    @cached_property
    def stdin(self) -> int | None:  # noqa: D102
        return None

    @cached_property
    def stdout(self) -> int | None:  # noqa: D102
        return self._config.output_pipes.out_handle

    @cached_property
    def stderr(self) -> int | None:  # noqa: D102
        return asyncio.subprocess.PIPE

    @cached_property
    def log_output(self) -> bool:  # noqa: D102
        return True

    @cached_property
    def log_stdout(self) -> bool:  # noqa: D102
        return False

    @cached_property
    def env(self) -> dict[str, str] | None:  # noqa: D102
        return None

    @cached_property
    def ignore_error(self) -> bool:  # noqa: D102
        # ld-dropout-correct does not support -l or -s flags and will
        # be killed when other procs are finished.
        # This is unfortunate as we would have to use more
        # complicated logic to determine a real crash from success.
        # Currently we do not check any errors from ld-dropout-correct
        # and assume all returncodes are success.
        return True

    @cached_property
    def stop_on_last_alive(self) -> bool:  # noqa: D102
        # On NT systems closing an os.pipe() does not kill the procs using
        # the pipe. The result of this is ld-dropout-correct continuing
        # to run after other procs have finished when using -l or -s.
        # This will signal to the proc killer to kill the process if it
        # is the only remaining proc type running.
        # This is essentially a workaround until ld-dropout-correct
        # supports -s/-l or properly supports named pipes.
        return True



from __future__ import annotations

import asyncio
import logging
import os
from typing import TYPE_CHECKING

from tbc_video_export.common import consts
from tbc_video_export.common.enums import ProcessName
from tbc_video_export.common.utils import ansi, log, strings

if TYPE_CHECKING:
    from tbc_video_export.process.parser.parser import Parser
    from tbc_video_export.process.process import Process
    from tbc_video_export.process.process_state import ProcessState
    from tbc_video_export.program_state import ProgramState as ProgramState


class ProgressHandler:
    """Handle the progress output.

    This requires ANSI escape code support in the terminal.
    """

    def __init__(
        self,
        state: ProgramState,
        procs: list[Process],
    ) -> None:
        self._state = state
        self._procs = procs
        self._state_str = str(self._state)

        self._col_w: dict[str, int] = {
            "sender": 22,
            "status": 2,
            "proc_name": 19,
            "tbc_type": 11,
            "errors": 3,
            "fps": 4,
            "tracked_name": 6,
            "tracked_value": 7,
            "size": 7,
            "duration": 12,
            "bitrate": 16,
        }

        log.setup_logger("progress", terminator="")

    async def print_progress_coroutine(self, interval: float = 0.1) -> None:
        """Print current progress.

        This loops through the process handles at the set interval.
        """
        with ansi.create_terminal_buffer():
            while not all(proc.state.is_stopped for proc in self._procs):
                self._print_progress()
                await asyncio.sleep(interval)

        # final print without headers
        self._print_progress(True)

    def _print_progress(self, final_print: bool = False) -> None:
        """Print progress lines.

        This will include the program header and state information unless final_print
        is set.
        """
        # number of procs that are running
        max_line_length = os.get_terminal_size().columns
        output_line = ""

        if not final_print:
            # reset terminal
            output_line += f"{ansi.move_to_home()}{ansi.erase_screen()}"

            # add header + state
            output_line += f"{strings.application_header()}\n\n{self._state_str}\n\n"

        # get new progress lines
        output_line += "".join(f"{self._progress(proc)}\n" for proc in self._procs)

        # add log and file data
        output_line += (
            f"{self._get_last_messages_line(max_line_length)}"
            f"{self._get_output_file_line()}"
        )

        if not final_print:
            output_line += f"\n{self._get_exit_help_line()}"

        logging.getLogger("progress").info(output_line)

    def _get_output_file_line(self) -> str:
        """Return a string containing stats on the finale output file."""
        return (
            f"\n"
            f"{ansi.dim('Size:')} "
            f"{self._state.export.size:>{self._col_w['size']}.2f} GB\t"
            f"{ansi.dim('Duration:')} "
            f"{self._state.export.duration[:-3]:>{self._col_w['duration']}s}\t"
            f"{ansi.dim('Bitrate:')} "
            f"{self._state.export.bitrate:>{self._col_w['bitrate']}s}"
            f"\n"
        )

    def _get_last_messages_line(self, max_line_length: int) -> str:
        """Get the last N lines sent from the procs.

        The amount of lines returned is based on the terminal height. We assume a
        minimum height of at least 30 and increase/decrease the number of log lines
        based on that.
        """
        terminal_height = os.get_terminal_size().lines
        extra_lines = terminal_height - consts.MINIMUM_TERMINAL_HEIGHT
        max_message_lines = 5 + extra_lines

        if max_message_lines > 0:
            messages = self._state.export.messages[-max_message_lines:]

            fixed_log_lines: list[str] = []

            # get the last N log lines, padding out the list if they do not exist
            for message in messages + [None] * (max_message_lines - len(messages)):
                if message is not None:
                    sender = (
                        f"[{message.process}]"
                        if message.process is not ProcessName.NONE
                        else f"[{consts.APPLICATION_NAME}]"
                    )

                    formatted_line = (
                        f"[{strings.formatted_timestamp(message.timestamp)}] "
                        f"{sender:<{self._col_w['sender']}s}{message.message}"
                    )

                    fixed_log_lines.append(
                        f"{formatted_line[:max_line_length - 3]}...\n"
                        if len(formatted_line) > max_line_length
                        else f"{formatted_line}\n"
                    )

            return "\n" + "".join(fixed_log_lines) + "\n"
        return ""

    def _progress(self, process: Process) -> str:
        """Get formatted progress string."""
        return (
            (
                self._status_icon(process.state)
                + self._formatted_process(process)
                + self._formatted_current(process.output_parser)
                + self._formatted_errors(process.output_parser)
                + self._formatted_fps(process.output_parser)
            )
            if process.state.is_running
            or any((process.state.is_errored, process.state.is_successful))
            else self._status_icon(process.state)
            + ansi.dim(self._formatted_process(process))
        )

    def _status_icon(self, process_state: ProcessState) -> str:
        """Get the status icon for the process.

        If the process is not in a success or error state, this will cycle through the
        list of running symbols to create an animation.
        """
        process_state.status_index = (
            process_state.status_index + 1
            if process_state.status_index < len(consts.RUNNING_SYMBOLS) - 1
            else 0
        )

        match process_state:
            case _ as state if state.is_successful:
                return ansi.success_color(
                    f"{consts.SUCCESS_SYMBOL:<{self._col_w['status']}s}"
                )

            case _ as state if state.is_errored:
                return ansi.error_color(
                    f"{consts.ERROR_SYMBOL:<{self._col_w['status']}s}"
                )

            case _ as state if state.is_running:
                return ansi.progress_color(
                    f"{consts.RUNNING_SYMBOLS[process_state.status_index]:<{self._col_w['status']}s}"
                )

            case _:
                return f"{' ' * self._col_w['status']}"

    def _formatted_process(self, process: Process) -> str:
        """Get formatted processs string."""
        value = f"{process.wrapper.process_name:<{self._col_w['proc_name']}s}"

        value += (
            f"{f'({process.wrapper.tbc_type})':<{self._col_w['tbc_type']}s}"
            if not process.output_parser.hide_tbc_type
            else (" " * self._col_w["tbc_type"])
        ) + " "

        return value

    def _formatted_errors(self, output: Parser) -> str:
        """Get formatted error string."""
        return f"{ansi.dim('errors:')} {output.error_count:>{self._col_w['errors']}d}  "

    def _formatted_fps(self, output: Parser) -> str:
        """Get formatted fps string."""
        return (
            f"{ansi.dim('fps:')} {output.current_fps:>{self._col_w['fps']}.0f}  "
            if output.current_fps > 0.0
            else ""
        )

    def _formatted_current(self, output: Parser) -> str:
        """Get formatted current string."""
        if output.tracked_value_name:
            value = ansi.dim(
                f"{output.tracked_value_name:>{self._col_w['tracked_name']}s}:"
            )

            value += f"{output.tracked_value:>{self._col_w['tracked_value']}d}"

            value += (
                f"/{output.tracked_value_total:<{self._col_w['tracked_value']}d}"
                if output.tracked_value_total > 0
                else (" " * (self._col_w["tracked_value"] + 1))
            ) + "  "
        else:
            value = " " * (
                self._col_w["tracked_name"] + (self._col_w["tracked_value"] * 2) + 4
            )

        return value

    @staticmethod
    def _get_exit_help_line() -> str:
        """Return line explaining how to exit the encode."""
        return ansi.dim("Press Ctrl+C to cancel the export.")



from __future__ import annotations

import asyncio
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING

from tbc_video_export.common import consts, exceptions
from tbc_video_export.common.enums import (
    ExportMode,
    HardwareAccelType,
    PipeType,
    ProcessName,
    TBCType,
    VideoFormatType,
    VideoSystem,
)
from tbc_video_export.common.utils import FlatList, ansi
from tbc_video_export.process.wrapper.wrapper import Wrapper

if TYPE_CHECKING:
    from tbc_video_export.config.profile import Profile, ProfileVideo
    from tbc_video_export.process.wrapper.pipe import Pipe
    from tbc_video_export.process.wrapper.wrapper import WrapperConfig
    from tbc_video_export.program_state import ProgramState


class WrapperFFmpeg(Wrapper):
    """Wrapper for ffmpeg that generates commands for encoding."""

    def __init__(
        self, state: ProgramState, config: WrapperConfig[tuple[Pipe], None]
    ) -> None:
        self._additional_vopts: FlatList = FlatList()
        self._hwaccel_opts: FlatList = FlatList()
        self._hwaccel_filter: str = ""
        super().__init__(state, config)
        self._config = config
        self._parse_hwaccel()

    def post_fn(self) -> None:  # noqa: D102
        pass

    @property
    def command(self) -> FlatList:  # noqa: D102
        return FlatList(
            (
                self.binary,
                self._get_verbosity_opts(),
                self._get_overwrite_opt(),
                self._get_threads_opt(),
                "-nostdin",
                self._get_hwaccel_opts(),
                self._get_color_range_opt(),
                self._get_input_opts(),
                self._get_filter_complex_opts(),
                self._get_map_opts(),
                self._get_timecode_opt(),
                self._get_framerate_opt(),
                self._get_color_range_opt(),
                self._get_color_opts(),
                self._get_codec_opts(),
                self._get_metadata_opts(),
                self._get_output_opt(),
            )
        )

    def _get_verbosity_opts(self) -> FlatList:
        """Return opts for verbosity."""
        verbosity_opts = FlatList("-hide_banner")

        # enable progress reporting if we are not displaying output
        if not self._state.opts.show_process_output:
            verbosity_opts.append(
                (
                    "-loglevel",
                    "error",
                    "-progress",
                    "pipe:2",
                )
            )
        else:
            verbosity_opts.append(("-loglevel", "verbose"))

        return verbosity_opts

    def _get_overwrite_opt(self) -> FlatList | None:
        """Return opts for overwrite enabled."""
        return FlatList(self._state.opts.convert_opt("overwrite", "-y"))

    def _get_threads_opt(self) -> FlatList:
        """Return opts for thread count."""
        # TODO add complex filter threads?
        return FlatList(("-threads", str(self._state.opts.threads)))

    def _parse_hwaccel(self) -> None:
        """Parse hardware acceleration opts."""
        match self._get_profile().video_profile.hardware_accel:
            case HardwareAccelType.VAAPI:
                self._hwaccel_opts.append(
                    ("-hwaccel", "vaapi", "-hwaccel_output_format", "vaapi")
                )

                if (hwaccel_device := self._state.opts.hwaccel_device) is not None:
                    self._hwaccel_opts.append(("-vaapi_device", hwaccel_device))

                self._hwaccel_filter = "hwupload"

            case HardwareAccelType.NVENC:
                if (hwaccel_device := self._state.opts.hwaccel_device) is not None:
                    self._additional_vopts.append(("-gpu", hwaccel_device))

            case HardwareAccelType.QUICKSYNC:
                self._hwaccel_opts.append(
                    (
                        "-hwaccel",
                        "qsv",
                        "-hwaccel_output_format",
                        "qsv",
                    )
                )

                if (hwaccel_device := self._state.opts.hwaccel_device) is not None:
                    self._hwaccel_opts.append(("-qsv_device", hwaccel_device))

            case HardwareAccelType.AMF:
                if self._state.opts.hwaccel_device is not None:
                    raise exceptions.InvalidProfileError(
                        "Unable to set device for AMD AMF encoding due to FFmpeg "
                        "limitiations."
                    )

            case _:
                pass

    def _get_hwaccel_opts(self) -> FlatList:
        return self._hwaccel_opts

    @staticmethod
    def _get_color_range_opt() -> FlatList | None:
        """Return opts for color range."""
        return FlatList(("-color_range", "tv"))

    def _get_thread_queue_size_opt(self) -> FlatList:
        """Return opts for thread queue size."""
        return FlatList(("-thread_queue_size", str(self._state.opts.thread_queue_size)))

    def _get_input_opts(self) -> FlatList:
        """Return opts for all inputs."""
        return FlatList(
            (
                self._get_video_input_opts(),
                self._get_audio_input_opts(),
                self._get_metadata_input_opts(),
            )
        )

    def _get_video_input_opts(self) -> FlatList:
        """Return opts for video input."""
        input_opts = FlatList()

        if self._config.export_mode == ExportMode.LUMA_4FSC:
            input_opts.append(
                (
                    "-f",
                    "rawvideo",
                    "-pix_fmt",
                    "gray16le",
                    "-framerate",
                    self._get_framerate(),
                    "-video_size",
                )
            )

            match self._state.video_system:
                case VideoSystem.PAL:
                    input_opts.append("1135x626")

                case VideoSystem.NTSC:
                    input_opts.append("910x526")

                case VideoSystem.PAL_M:
                    input_opts.append("909x526")

        if self._is_two_step_merge_mode():
            # add the luma file as an input
            input_opts.append(
                (
                    self._get_thread_queue_size_opt(),
                    "-i",
                    self._state.file_helper.output_video_file_luma,
                )
            )

        # pipes
        input_opts.append(
            (self._get_thread_queue_size_opt(), "-i", str(i.in_path))
            for i in self._config.input_pipes
        )

        return input_opts

    def _get_audio_input_opts(self) -> FlatList:
        """Return opts for audio input."""
        input_opts = FlatList()

        # audio
        for track in self._state.opts.audio_track:
            if (offset := track.offset) is not None:
                input_opts.append(("-itsoffset", offset))

            if (sample_format := track.sample_format) is not None:
                input_opts.append(("-f", sample_format))

            if (rate := track.rate) is not None:
                input_opts.append(("-ar", rate))

            if (channels := track.channels) is not None:
                input_opts.append(("-ac", channels))

            input_opts.append(("-i", track.file_name))

            # if the track does not exist, we add a warning to the message log as
            # the file may be generated by ld-process-efm and will exist by the time
            # FFmpeg runs
            # we should perhaps only init a wrapper when it is time to run?
            if not Path(track.file_name).is_file():
                self._state.export.append_message(
                    ansi.error_color(
                        f"FFmpeg track {track.file_name} does not currently exist."
                    )
                )

        return input_opts

    def _get_metadata_input_opts(self) -> FlatList:
        """Return opts for metadata input."""
        input_opts = FlatList()
        if self._state.opts.export_metadata:
            # subtitles
            if self._state.opts.dry_run:
                input_opts.append(("{-i", "[SUBTITLE_FILE]}"))
            elif (cc_file := self._state.file_helper.cc_file).is_file():
                input_opts.append(("-i", cc_file))

            # metadata
            if self._state.opts.dry_run:
                input_opts.append(("{-i", "[METADATA_FILE]}"))
            elif (ffmetadata := self._state.file_helper.ffmetadata_file).is_file():
                input_opts.append(("-i", ffmetadata))

        for ffmetadata in self._state.opts.metadata_file:
            input_opts.append(("-i", ffmetadata))

        return input_opts

    def _get_vbi_crop_filter(self) -> str:
        """Return crop filter for full vertical to VBI crop."""
        match self._state.video_system:
            case VideoSystem.PAL:
                return "crop=iw:ih-12:0:9"

            case VideoSystem.NTSC | VideoSystem.PAL_M:
                return "crop=iw:ih-19:0:17"

    def _get_filters(self) -> tuple[list[str], list[str]]:
        """Return tuple containing video and other filters."""
        video_filters: list[str] = []
        other_filters: list[str] = []

        # add full vertical -> vbi crop
        if self._state.opts.vbi or self._get_profile().include_vbi:
            video_filters.append(self._get_vbi_crop_filter())

        _vf, _of = self._state.config.get_profile_filters(self._get_profile())

        # set video filters
        video_filters += _vf

        if self._state.opts.force_anamorphic or self._state.opts.letterbox:
            video_filters.append(self._get_widescreen_aspect_ratio_filter())

        # override profile colorlevels if set with opt
        if self._state.opts.force_black_level is not None:
            video_filters.append(
                "colorlevels="
                f"rimin={self._state.opts.force_black_level[0]}/255:"
                f"gimin={self._state.opts.force_black_level[1]}/255:"
                f"bimin={self._state.opts.force_black_level[2]}/255"
            )

        if self._state.opts.append_video_filter is not None:
            video_filters.append(self._state.opts.append_video_filter)

        video_filters.append(f"format={self._get_profile_video_format()}")

        if self._state.opts.hwaccel_type is not None and self._hwaccel_filter:
            video_filters.append(self._hwaccel_filter)

        # set other filters
        other_filters += _of

        if self._state.opts.append_other_filter is not None:
            other_filters.append(self._state.opts.append_other_filter)

        return video_filters, other_filters

    def _get_filter_complex_opts(self) -> FlatList:  # noqa: C901, PLR0912
        """Return opts for filter complex."""
        field_filter = f"setfield={self._get_field_order()}"
        video_filters, other_filters = self._get_filters()

        # add setfield to start of filters
        video_filters.insert(0, field_filter)

        video_filters_opts = ",".join(video_filters)

        other_filters_str = ",".join(other_filters)
        other_filters_opts = f",{other_filters_str}" if len(other_filters_str) else ""

        match self._config.export_mode:
            case ExportMode.CHROMA_MERGE:
                # merge Y/C from separate Y+C inputs

                # using mergeplanes 0x001112 with pipe+pipe input works but there
                # seems to be an issue when merging gray16le and 16-bit yuv(??) formats.
                # safer to extract the 2x inputs (file+pipe/pipe+pipe) into y/u/v planes
                # and merge to avoid any issues.

                mergeplanes = (
                    "0x001020"
                    if consts.FFMPEG_USE_OLD_MERGEPLANES
                    else "map1s=1:map2s=2"
                )

                complex_filter = (
                    f"[0:v]format=gray16le[luma];[1:v]format=yuv444p16le[chroma];"
                    f"[luma]extractplanes=y[y];[chroma]extractplanes=u+v[u][v];"
                    f"[y][u][v]mergeplanes={mergeplanes}:format=yuv444p16le,"
                    f"{video_filters_opts}[v_output]"
                    f"{other_filters_opts}"
                )

            case ExportMode.LUMA_EXTRACTED:
                # extract Y from a Y/C input
                complex_filter = (
                    f"[0:v]extractplanes=y,{video_filters_opts}"
                    f"[v_output]"
                    f"{other_filters_opts}"
                )

            case ExportMode.LUMA_4FSC:
                # interleve tbc fields
                complex_filter = (
                    f"[0:v]il=l=i:c=i,{video_filters_opts}"
                    f"[v_output]"
                    f"{other_filters_opts}"
                )

            case _ as mode if mode is ExportMode.LUMA and self._state.opts.two_step:
                # luma step in two-step should not use any filters (excluding setfield)
                complex_filter = f"[0:v]{field_filter}[v_output]"

            case _:
                complex_filter = (
                    f"[0:v]{video_filters_opts}[v_output]{other_filters_opts}"
                )

        return FlatList(("-filter_complex", complex_filter))

    def _get_map_opts(self) -> FlatList:
        """Return FFmpeg video map opts."""
        # video
        input_opts = FlatList(("-map", "[v_output]"))
        input_count = len(self._config.input_pipes)

        # audio
        input_opts.append(
            ("-map", f"{i + input_count}:a")
            for i in range(len(self._state.opts.audio_track))
        )
        input_count += len(self._state.opts.audio_track)

        if self._state.opts.export_metadata:
            # subtitles
            if self._state.opts.dry_run:
                input_opts.append(("{-map", "[SUBTITLE_INDEX]:s}"))
            elif self._state.file_helper.cc_file.is_file():
                input_opts.append(("-map", f"{input_count}:s"))
                input_count += 1

            # metadata
            if self._state.opts.dry_run:
                input_opts.append(("{-map_metadata", "[METADATA_INDEX]}"))
            elif self._state.file_helper.ffmetadata_file.is_file():
                input_opts.append(("-map_metadata", input_count))
                input_count += 1

        for _ in self._state.opts.metadata_file:
            input_opts.append(("-map_metadata", input_count))
            input_count += 1

        return input_opts

    def _get_timecode_opt(self) -> FlatList:
        """Return opts for timecode."""
        return FlatList(("-timecode", self._state.file_helper.tbc_json.timecode))

    def _get_framerate(self) -> str:
        """Return rate based on video system."""
        match self._state.video_system:
            case VideoSystem.PAL:
                return "pal"

            case VideoSystem.NTSC | VideoSystem.PAL_M:
                return "ntsc"

    def _get_framerate_opt(self) -> FlatList:
        """Return opts for rate."""
        return FlatList(
            (
                "-framerate",
                self._get_framerate(),
            )
        )

    def _get_widescreen_aspect_ratio_filter(self) -> str:
        """Return filter for widescreen aspect ratio."""
        match self._state.video_system:
            case VideoSystem.PAL:
                return "setsar=865/779:max=1000"

            case VideoSystem.NTSC | VideoSystem.PAL_M:
                return "setsar=25/22:max=1000"

    def _get_color_opts(self) -> FlatList | None:
        """Return opts for color settings."""
        match self._state.video_system:
            case VideoSystem.PAL | VideoSystem.PAL_M:
                return FlatList(
                    (
                        "-colorspace",
                        "bt470bg",
                        "-color_primaries",
                        "bt470bg",
                        "-color_trc",
                        "bt709",
                    ),
                )
            case VideoSystem.NTSC:
                return FlatList(
                    (
                        "-colorspace",
                        "smpte170m",
                        "-color_primaries",
                        "smpte170m",
                        "-color_trc",
                        "bt709",
                    ),
                )

    def _get_codec_opts(self) -> FlatList:
        """Return opts containing codecs for inputs."""
        codec_opts = FlatList(
            (
                "-c:v",
                self._get_profile().video_profile.codec,
                self._get_profile().video_profile.opts,
                self._additional_vopts,
            )
        )

        if (audio_profile := self._get_profile().audio_profile) is not None:
            codec_opts.append(
                (
                    "-c:a",
                    audio_profile.codec,
                    audio_profile.opts,
                )
            )

            if self._state.opts.export_metadata:
                if self._state.opts.dry_run:
                    codec_opts.append(("{-c:s", self._get_subtitle_format() + "}"))
                elif (
                    self._state.opts.export_metadata
                    and self._state.file_helper.cc_file.is_file()
                ):
                    codec_opts.append(("-c:s", self._get_subtitle_format()))

        return codec_opts

    def _get_metadata_opts(self) -> FlatList:
        """Return opts for metadata."""
        # custom metadata
        metadata_opts = FlatList(
            ("-metadata", f"{data[0]}={data[1]}") for data in self._state.opts.metadata
        )

        # audio
        for idx, track in enumerate(self._state.opts.audio_track):
            if (title := track.title) is not None:
                metadata_opts.append((f"-metadata:s:a:{idx}", f"title={title}"))

            if (language := track.language) is not None:
                metadata_opts.append((f"-metadata:s:a:{idx}", f"language={language}"))

            if (layout := track.layout) is not None:
                metadata_opts.append((f"-channel_layout:a:{idx}", f"{layout}"))

        # attachment
        if self._get_supports_attachments() and not self._state.opts.no_attach_json:
            metadata_opts.append(
                (
                    "-attach",
                    self._state.file_helper.tbc_json.file_name,
                    "-metadata:s:t:0",
                    "mimetype=application/json",
                )
            )

        return metadata_opts

    def _get_output_opt(self) -> FlatList:
        """Output opts for ffmpeg."""
        output_file = (
            self._state.file_helper.output_video_file_luma
            if self._is_two_step_luma_mode()
            else self._state.file_helper.output_video_file
        )

        # only set if the user has not manually specified a container
        output_format = (
            self._get_profile().video_profile.output_format
            if not self._state.opts.profile_container
            else None
        )

        if self._state.opts.checksum:
            checksum_output = (
                f"[f={output_format}]{output_file}"
                if output_format is not None
                else output_file
            )

            return FlatList(
                (
                    "-f",
                    "tee",
                    f"[f=streamhash]{output_file}.sha256|{checksum_output}",
                )
            )

        return (
            FlatList(("-f", output_format, output_file))
            if output_format is not None
            else FlatList(output_file)
        )

    def _get_profile(self) -> Profile:
        """Return the profile in state."""
        return self._state.profile

    def _get_video_profile(self) -> ProfileVideo:
        """Return the video profile in state."""
        return self._state.profile.video_profile

    def _get_profile_video_format(self) -> str:
        """Return the video format in state."""
        video_format = self._get_profile().video_profile.video_format

        # if two step, set to gray8/16le
        if self._is_two_step_luma_mode() or self._config.export_mode in (
            ExportMode.LUMA,
            ExportMode.LUMA_4FSC,
            ExportMode.LUMA_EXTRACTED,
        ):
            depth = (
                16
                if self._state.opts.video_bitdepth is None
                else self._state.opts.video_bitdepth
            )

            if (new_format := VideoFormatType.GRAY.value.get(depth)) is not None:
                video_format = new_format

        # check bitdepth override
        if (depth := self._state.opts.video_bitdepth) is not None and (
            new_format := VideoFormatType.get_new_format(video_format, depth)
        ) is not None:
            video_format = new_format

        # if override opts, ensure format for bitdepth and set
        # does not set the luma format when in two-step mode
        if (
            (vf := self._state.opts.video_format) is not None
            and (depth := self._state.opts.video_bitdepth) is not None
            and (new_format := vf.value.get(depth)) is not None
            and not self._is_two_step_luma_mode()
        ):
            video_format = new_format

        return video_format

    def _is_two_step_luma_mode(self) -> bool:
        """Return True if this wrapper is in luma mode while two-step is enabled."""
        return self._state.opts.two_step and self._config.export_mode is ExportMode.LUMA

    def _is_two_step_merge_mode(self) -> bool:
        """Return True if this wrapper is in merge mode while two-step is enabled."""
        return (
            self._state.opts.two_step
            and self._config.export_mode is ExportMode.CHROMA_MERGE
        )

    def _get_supports_attachments(self) -> bool:
        """Return True if attachments are supported by the profile container."""
        return self._state.file_helper.output_container.lower() == "mkv"

    def _get_subtitle_format(self) -> str:
        """Return subtitle format based on the profile container."""
        return (
            "srt"
            if self._state.file_helper.output_container.lower() != "mov"
            else "mov_text"
        )

    def _get_field_order(self) -> str:
        """Return the formatted field order from opts."""
        return self._state.opts.field_order.name.lower()

    @cached_property
    def process_name(self) -> ProcessName:  # noqa: D102
        return ProcessName.FFMPEG

    @cached_property
    def supported_pipe_types(self) -> PipeType:  # noqa: D102
        return PipeType.NULL | PipeType.OS | PipeType.NAMED

    @cached_property
    def stdin(self) -> int | None:  # noqa: D102
        # if ffmpeg has a pipe with a handle, use it
        # we usually use named pipes that do not have handles
        return next(
            (
                pipe.in_handle
                for pipe in self._config.input_pipes
                if pipe.in_handle is not None
            ),
            None,
        )

    @cached_property
    def stdout(self) -> int | None:  # noqa: D102
        # if ffmpeg has a pipe with a handle, use it
        # we usually do not have any out pipes
        return next(
            (
                pipe.in_handle
                for pipe in self._config.input_pipes
                if pipe.in_handle is not None
            ),
            None,
        )

    @cached_property
    def stderr(self) -> int | None:  # noqa: D102
        return asyncio.subprocess.PIPE

    @cached_property
    def log_output(self) -> bool:  # noqa: D102
        # we use built in FFmpeg logging via env
        return False

    @cached_property
    def log_stdout(self) -> bool:  # noqa: D102
        return False

    @cached_property
    def env(self) -> dict[str, str] | None:  # noqa: D102
        if self._state.opts.log_process_output:
            file_name = self._state.file_helper.get_log_file(
                self.process_name, TBCType.NONE
            )

            return {"FFREPORT": f"file='{file_name}'"}
        return None

    @cached_property
    def ignore_error(self) -> bool:  # noqa: D102
        return False

    @cached_property
    def stop_on_last_alive(self) -> bool:  # noqa: D102
        return False



from __future__ import annotations

import os
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING

from tbc_video_export.common import consts, exceptions
from tbc_video_export.common.enums import FlagHelper, ProcessName, TBCType
from tbc_video_export.common.tbc_json_helper import TBCJsonHelper
from tbc_video_export.common.utils import files
from tbc_video_export.config.config import GetProfileFilter

if TYPE_CHECKING:
    from tbc_video_export.config.config import Config
    from tbc_video_export.opts import Opts


class FileHelper:
    """Helper for files.

    Handles reading the TBC json, checking if files exist and generating paths.
    """

    def __init__(self, opts: Opts, config: Config) -> None:
        self._opts = opts
        self._config = config

        self._profile = self._config.get_profile(GetProfileFilter(self._opts.profile))

        # initially set both input and output files to the input file
        # file without file extension
        self._input_path = self._output_path = Path(self._opts.input_file).parent

        # path to file
        self._input_file_name = self._output_file_name = Path(
            self._opts.input_file
        ).stem

        # set output file path/name if set to new path
        if self._opts.output_file is not None:
            self._output_path = Path(self._opts.output_file).parent
            self._output_file_name = Path(self._opts.output_file).stem

        self.tools = self._get_tool_paths()
        self.tbcs = self._set_tbc_files()

    @cached_property
    def input_name(self) -> Path:
        """Return absolute path to input file without file extension."""
        return Path(self._input_path).joinpath(self._input_file_name)

    @cached_property
    def output_name(self) -> Path:
        """Return absolute path to output without file extension."""
        return Path(self._output_path).joinpath(self._output_file_name)

    @cached_property
    def is_combined_ld(self) -> bool:
        """Returns True if the TBC is for LaserDisc.

        This only checks if the TBC type is combined and an EFM file is
        located. There may be a more reliable way of doing this.
        """
        return TBCType.COMBINED in self.tbcs and self.efm_file is not None

    @property
    def efm_file(self) -> Path | None:
        """Returns absolute path to EFM file if it exists."""
        if (file := self.get_input_file_from_ext("efm")).is_file():
            return file
        return None

    @property
    def ffmetadata_file(self) -> Path:
        """Returns absolute path fo metadata file."""
        return self.get_output_file_from_ext("ffmetadata")

    @property
    def cc_file(self) -> Path:
        """Returns absolute path to subtitle file ."""
        return self.get_output_file_from_ext("scc")

    @property
    def tbc_types(self) -> TBCType:
        """Returns all TBC types found."""
        types = TBCType.NONE

        for t in self.tbcs:
            types |= t

        # remove none if others set
        if types is not TBCType.NONE:
            types &= ~TBCType.NONE

        return types

    @property
    def tbc_json(self) -> TBCJsonHelper:
        """Returns TBCJson helper.

        This will create the helper if it does not exist.
        """
        if not getattr(
            self,
            "_tbc_json",
            False,
        ):
            self._tbc_json = TBCJsonHelper(
                Path(self._opts.input_tbc_json)
                if self._opts.input_tbc_json is not None
                else Path(f"{self.input_name}.tbc.json")
            )

        return self._tbc_json

    @tbc_json.setter
    def tbc_json(self, file_name: Path) -> None:
        """Set the TBCJson Helper.

        This can be used when ld-process-vbi generates a new JSON file.
        """
        self._tbc_json = TBCJsonHelper(file_name)

    @cached_property
    def tbc_luma(self) -> Path:
        """Return the absolute path to the luma (or combined) TBC."""
        if TBCType.LUMA in self.tbcs:
            return self.tbcs[TBCType.LUMA]

        if TBCType.COMBINED in self.tbcs:
            return self.tbcs[TBCType.COMBINED]

        raise exceptions.TBCError("Unable to find luma TBC.")

    @cached_property
    def output_container(self) -> str:
        """Return container used for output video file."""
        return (
            self._opts.profile_container
            if self._opts.profile_container is not None
            else self._profile.video_profile.container
        )

    @cached_property
    def output_video_file(self) -> Path:
        """Return absolute path to the output video file."""
        return self.get_output_file_from_ext(self.output_container)

    @cached_property
    def output_video_file_luma(self) -> Path:
        """Return absolute path to the output video file for luma.

        This is used when two-step is enabled when merging.
        """
        return self.get_output_file_from_ext(
            f"{consts.TWO_STEP_OUT_FILE_LUMA_SUFFIX}.{self.output_container}"
        )

    def get_log_file(
        self,
        process_name: ProcessName,
        tbc_type: TBCType,
        timestamp: str = consts.CURRENT_TIMESTAMP,
    ):
        """Return absolute path to log file for process/tbc type."""
        return Path(self._output_path).joinpath(
            f"{timestamp}_{self._input_file_name}_{process_name}"
            f"_{FlagHelper.get_flags_str(tbc_type, '_')}.log"
        )

    def get_output_file_from_ext(self, extension: Path | str) -> Path:
        """Return absolute path to output file with extension."""
        return Path(f"{self.output_name}.{extension}")

    def get_input_file_from_ext(self, extension: Path | str) -> Path:
        """Return absolute path to input file with extension."""
        return Path(f"{self.input_name}.{extension}")

    def get_tool(self, process_name: ProcessName) -> Path | list[Path | str]:
        """Return absolute path to tool binary from process."""
        # if appimage is set, all binary paths are set to the appimage location
        # so we return a list containing the appimage file with the binary
        # name we want to use
        if self._opts.appimage:
            return [self.tools[process_name], str(process_name)]

        return self.tools[process_name]

    def check_output_dir(self) -> None:
        """Check if output directory exists.

        This throws an exception if output directory does not exist.
        """
        if not Path(self._output_path).is_dir():
            raise exceptions.FileIOError(
                f"Output directory does not exist ({self._output_path})."
            )

    def check_output_file(self) -> None:
        """Check if output file exists.

        This throws an exception if output file exists.
        """
        if not self._opts.overwrite:
            files = [self.output_video_file]

            if self._opts.two_step:
                files.append(self.output_video_file_luma)

            for file in files:
                if file.is_file():
                    raise exceptions.FileIOError(
                        f"{file} exists, use --overwrite or move the file."
                    )

    def _set_tbc_files(self) -> dict[TBCType, Path]:
        """Create a dict containing the absolute path to the TBC files based on type."""
        tbcs: dict[TBCType, Path] = {}

        # input files
        tbc = f"{self.input_name}.tbc"
        tbc_chroma = f"{self.input_name}_chroma.tbc"

        if (tbc_chroma := Path(tbc_chroma)).is_file():
            tbcs[TBCType.CHROMA] = tbc_chroma

        if (tbc := Path(tbc)).is_file():
            if TBCType.CHROMA in tbcs:
                tbcs[TBCType.LUMA] = tbc
            else:
                tbcs[TBCType.COMBINED] = tbc

        # ensure tbcs exist
        if len(tbcs) == 0:
            raise exceptions.TBCError("TBC not found at location.")

        # if for some reason we found a _chroma.tbc but no .tbc
        if TBCType.CHROMA in tbcs and TBCType.LUMA not in tbcs:
            raise exceptions.TBCError("Location contains chroma TBC but no luma TBC.")

        return tbcs

    def _get_tool_paths(self) -> dict[ProcessName, Path]:
        """Get required tool paths from PATH or script path."""
        tools: dict[ProcessName, Path] = {}
        tools[ProcessName.FFMPEG] = self._get_tool_path(ProcessName.FFMPEG)

        if not self._opts.no_dropout_correct and not self._opts.luma_4fsc:
            tools[ProcessName.LD_DROPOUT_CORRECT] = self._get_tool_path(
                ProcessName.LD_DROPOUT_CORRECT
            )

        if not self._opts.luma_4fsc:
            tools[ProcessName.LD_CHROMA_DECODER] = self._get_tool_path(
                ProcessName.LD_CHROMA_DECODER
            )

        if self._opts.process_vbi:
            tools[ProcessName.LD_PROCESS_VBI] = self._get_tool_path(
                ProcessName.LD_PROCESS_VBI
            )

        if self._opts.process_efm:
            tools[ProcessName.LD_PROCESS_EFM] = self._get_tool_path(
                ProcessName.LD_PROCESS_EFM
            )

        if self._opts.export_metadata:
            tools[ProcessName.LD_EXPORT_METADATA] = self._get_tool_path(
                ProcessName.LD_EXPORT_METADATA
            )

        return tools

    def _get_tool_path(self, tool_name: ProcessName) -> Path:
        # if appimage is defined, use it as the location for the tools
        if os.name != "nt":
            binary = (
                self._opts.appimage
                if self._opts.appimage
                else tool_name.name.lower().replace("_", "-")
            )
        else:
            # append .exe on NT
            binary = f"{tool_name}.exe"

        return files.find_binary(binary)



from __future__ import annotations

import os
import sys
from functools import cache
from pathlib import Path
from shutil import which

from tbc_video_export.common import exceptions


@cache
def get_runtime_directory() -> Path:
    """Return the runtime directory.

    When the script is built to a single executable using PyInstaller __file__ is
    somewhere in TEMP, so the executable location must be used instead.
    """
    return (
        Path(sys.executable)
        if getattr(sys, "frozen", False)
        else Path(os.path.realpath(sys.argv[0]))
    )


@cache
def find_binary(name: str) -> Path:
    """Return the path of a binary if found.

    This searches in the same location as the script, the current dir
    and in PATH.
    """
    # check if binary exists at path provided
    if (path := Path(name)).is_file():
        return path

    # check if binary exists in the same dir as script
    script_path = get_runtime_directory().with_name(name).absolute()

    if Path(script_path).is_file():
        return script_path

    # check if binary exists in PATH or current dir
    if which(name):
        return Path(name)

    raise exceptions.FileIOError(f"{name} not in PATH or script dir.")



from __future__ import annotations

import re
from contextlib import suppress
from typing import TYPE_CHECKING

from tbc_video_export.process.parser.export_state import ExportStateSnapshot
from tbc_video_export.process.parser.parser import Parser

if TYPE_CHECKING:
    from tbc_video_export.common.enums import ProcessName


class ParserFFmpeg(Parser):
    """Parser for FFmpeg process."""

    def __init__(self, process_name: ProcessName) -> None:
        super().__init__(process_name)

        self.tracked_value_name = "frame"

    def parse_line(self, line: str) -> ExportStateSnapshot:  # noqa: D102
        state = ExportStateSnapshot()

        patterns = [
            r"frame=(.*?)$",
            r"fps=(.*?)$",
            r"bitrate=(.*?)$",
            r"total_size=(.*?)$",
            r"dup_frames=(.*?)$",
            r"drop_frames=(.*?)$",
            r"out_time=(.*?)$",
        ]

        with suppress(ValueError):
            if (reg := re.match("|".join(patterns), line)) is not None:
                # matches but no groups, an error
                # we should never see any other messages here, so exit
                if (frame := reg.group(1)) is not None:
                    self.tracked_value = int(frame)
                elif (fps := reg.group(2)) is not None:
                    self.current_fps = float(fps)
                elif (bitrate := reg.group(3)) is not None:
                    state.bitrate = bitrate
                elif (size := reg.group(4)) is not None:
                    state.size = int(size) / 1024 / 1024 / 1024
                elif None not in {reg.group(5), reg.group(6)}:
                    self.error_count += 1
                elif (duration := reg.group(7)) is not None:
                    state.duration = duration

        return state

    @property
    def hide_tbc_type(self) -> bool:  # noqa: D102
        return True



from __future__ import annotations

import asyncio
import signal
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any

    from tbc_video_export.process.process_handler import ProcessHandler


class InterruptHandler:
    """Interrupt handler context.

    Creates a context that creates a signal handler and releases it on exit.
    """

    def __init__(
        self, process_handler: ProcessHandler, signal: signal.Signals = signal.SIGINT
    ):
        self._process_handler = process_handler
        self._signal = signal
        self._released = False

        self._tasks: set[asyncio.Task[Any]] = set()

    def __enter__(self):
        """Enter the interrupt context.

        This creates a signal handler while in the conext.
        """
        self._original_handler = signal.getsignal(self._signal)
        signal.signal(self._signal, self._signal_handler)

        return self

    def __exit__(self, *_: object):
        """Exit the interrupt context.

        This releases the signal handler.
        """
        self._release_signal()

    def _signal_handler(self, *_: Any) -> None:
        """Triggered on signal.

        Creates a task to stop the application and release the signal.
        """
        self._tasks.add(asyncio.create_task(self._process_handler.stop(True)))
        self._release_signal()

        asyncio.gather(*self._tasks)

    def _release_signal(self) -> None:
        """Set the signal to the original handler."""
        if self._released:
            return

        signal.signal(self._signal, self._original_handler)
        self._released = True



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from tbc_video_export.common import consts
from tbc_video_export.common.utils import ansi

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Any

    from tbc_video_export.opts.opts import Opts


class ColorFormatter(logging.Formatter):
    """Formatter class for logging."""

    def format(self, record: logging.LogRecord):  # noqa: A003
        """Return colored formatter based on log level."""
        match record.levelno:
            case logging.DEBUG:
                formatter = logging.Formatter(ansi.dim_style("%(message)s"))

            case logging.INFO:
                formatter = logging.Formatter(ansi.default_color("%(message)s"))

            case logging.WARNING:
                formatter = logging.Formatter(ansi.default_color("%(message)s"))

            case logging.ERROR:
                formatter = logging.Formatter(ansi.error_color("%(message)s"))

            case logging.CRITICAL:
                formatter = logging.Formatter(ansi.error_color("%(message)s"))

            case _:
                formatter = logging.Formatter("%(message)s")

        return formatter.format(record)

    def formatException(self, ei: Any) -> str:  # noqa: N802
        """Return colored exception formatter."""
        return ansi.error_color(super().formatException(ei))


def setup_logger(
    name: str,
    enable_console: bool = True,
    filename: str | None = None,
    terminator: str = "\n",
) -> logging.Logger:
    """Set up logger."""
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)

    if enable_console:
        add_console_handler(logger, terminator=terminator)

    if filename is not None:
        add_debug_file_handler(logger, filename)

    return logger


def add_debug_file_handler(logger: logging.Logger, filename: str | Path) -> None:
    """Add a file handler to a logger."""
    file_handler = logging.FileHandler(filename)
    file_handler.setLevel(logging.DEBUG)

    logger.addHandler(file_handler)


def add_console_handler(
    logger: logging.Logger, level: int = logging.DEBUG, terminator: str = "\n"
) -> None:
    """Add a console handler to a logger."""
    console_handler = logging.StreamHandler()
    console_handler.setLevel(level)
    console_handler.setFormatter(ColorFormatter())
    console_handler.terminator = terminator
    logger.addHandler(console_handler)


def set_verbosity(opts: Opts) -> None:
    """Set verbosity and handlers based on opts."""
    if opts.quiet:
        logging.getLogger("console").setLevel(logging.ERROR)
        opts.no_progress = True
        opts.show_process_output = False
    elif opts.debug:
        # remove existing handlers and set base level
        logging.getLogger("console").handlers.clear()
        logging.getLogger("console").setLevel(logging.DEBUG)

        if opts.no_progress:
            # add new console handler with DEBUG level
            add_console_handler(logging.getLogger("console"))
        else:
            # re-add console handler with INFO level
            add_console_handler(logging.getLogger("console"), level=logging.INFO)

        if not opts.no_debug_log:
            add_debug_file_handler(
                logging.getLogger("console"),
                f"{consts.CURRENT_TIMESTAMP}_debug.log",
            )

    if opts.show_process_output:
        opts.no_progress = True



from __future__ import annotations

import asyncio
import logging
import os
from contextlib import suppress
from dataclasses import dataclass
from functools import cached_property
from typing import TYPE_CHECKING

from tbc_video_export.common import consts, exceptions
from tbc_video_export.common.enums import PipeType
from tbc_video_export.common.utils import strings
from tbc_video_export.process.wrapper.pipe.pipe import Pipe

if TYPE_CHECKING:
    from pathlib import Path
    from types import TracebackType

    from tbc_video_export.common.enums import ProcessName, TBCType

assert os.name == "nt"

import pywintypes  # noqa: E402
import win32file  # noqa: E402
import win32pipe  # noqa: E402
import win32security  # noqa: E402
import winerror  # noqa: E402

from tbc_video_export.common.utils import win32  # noqa: E402


class PipeNamedNT(Pipe):
    """Named pipe helper for NT systems.

    This uses synchronous I/O, because of this most win32 calls are
    blocking and must run in their own thread. Not much error checking
    is done here.

    I hate Windows named pipes.
    """

    @dataclass
    class PipeData:
        """Storage class for pipe data."""

        pipe_name: str | Path
        open_mode: int | Path
        pipe_mode = (
            win32pipe.PIPE_TYPE_BYTE
            | win32pipe.PIPE_READMODE_BYTE
            | win32pipe.PIPE_WAIT
        )
        max_instances = consts.NT_NAMED_PIPE_MAX_INSTANCES
        out_buffer_size = consts.NT_NAMED_PIPE_BUFFER_SIZE
        in_buffer_size = consts.NT_NAMED_PIPE_BUFFER_SIZE
        time_out = consts.NT_NAMED_PIPE_TIMEOUT
        security_attributes = win32security.SECURITY_ATTRIBUTES()

        thread: asyncio.Future[None] | None = None
        file_handle: int | None = None
        ready_event = asyncio.Event()

    def __init__(self, process_name: ProcessName, tbc_type: TBCType) -> None:
        super().__init__(process_name, tbc_type)
        self._bridge_thread: asyncio.Task[None] | None = None
        self._start_bridge_task: asyncio.Future[None] | None = None

        self._id = strings.random_characters(7)

        self._pipe_data_stdin: PipeNamedNT.PipeData = PipeNamedNT.PipeData(
            self.in_path, win32pipe.PIPE_ACCESS_OUTBOUND
        )
        self._pipe_data_stdout: PipeNamedNT.PipeData = PipeNamedNT.PipeData(
            self.out_path, win32pipe.PIPE_ACCESS_INBOUND
        )

    async def __aenter__(self) -> Pipe:
        """Enter named NT pipe context."""
        # starts the pipe bridge thread
        self._start_bridge_task = asyncio.create_task(self._start_pipe_bridge())

        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None | bool:
        """Exit named NT pipe context."""
        self.close()

    @cached_property
    def pipe_type(self) -> PipeType:  # noqa: D102
        return PipeType.NAMED_NT

    @cached_property
    def in_path(self) -> Path | str:  # noqa: D102
        return (
            rf"\\.\pipe\{consts.APPLICATION_NAME}-{self._id}-"
            rf"{self._process_name}-{self._tbc_type}-in"
        )

    @cached_property
    def out_path(self) -> Path | str:  # noqa: D102
        return (
            rf"\\.\pipe\{consts.APPLICATION_NAME}-{self._id}-"
            rf"{self._process_name}-{self._tbc_type}-out"
        )

    @property
    def in_handle(self) -> int | None:  # noqa: D102
        return None

    @property
    def out_handle(self) -> int | None:  # noqa: D102
        return None

    def close(self) -> None:
        """Close named pipe."""
        self._close_pipe(str(self.in_path), win32file.GENERIC_READ)
        self._close_pipe(str(self.out_path), win32file.GENERIC_WRITE)

    async def _start_pipe_bridge(self) -> None:
        """Start the pipe bridge.

        This reads from one named pipe and writes into another.
        """
        loop = asyncio.get_event_loop()

        # create pipes
        self._create_pipe(self._pipe_data_stdin)
        self._create_pipe(self._pipe_data_stdout)

        # connect to pipes
        # we run these in a thread as CreateNamedPipe is blocking
        self._pipe_data_stdin.thread = loop.run_in_executor(
            None, self._connect_pipe, self._pipe_data_stdin
        )
        self._pipe_data_stdout.thread = loop.run_in_executor(
            None, self._connect_pipe, self._pipe_data_stdout
        )

        await asyncio.gather(
            *[self._pipe_data_stdin.thread, self._pipe_data_stdout.thread]
        )

        loop.run_in_executor(None, self._pipe_bridge)

    def _pipe_bridge(self) -> None:
        """Pipe bridge thread.

        This is started once the pipes have been created and connected, and the pipe
        handles exist.
        """
        logging.getLogger("console").debug(
            f"Pipe bridge for {self._pipe_data_stdout.pipe_name} -> "
            f"{self._pipe_data_stdin.pipe_name} started"
        )

        if (
            self._pipe_data_stdin.file_handle is None
            or self._pipe_data_stdout.file_handle is None
        ):
            raise exceptions.PipeError("Pipe handles do not exist.")

        try:
            while True:
                # read data from stdout pipe
                read_hr, read_buf = win32file.ReadFile(
                    self._pipe_data_stdout.file_handle,
                    consts.NT_NAMED_PIPE_BUFFER_SIZE,
                )

                if read_hr != winerror.ERROR_SUCCESS:
                    self._handle_winerror(read_hr)

                # write data to stdin pipe
                write_hr, written_count = win32file.WriteFile(
                    self._pipe_data_stdin.file_handle, read_buf
                )

                if write_hr != winerror.ERROR_SUCCESS:
                    self._handle_winerror(write_hr)

                if written_count != len(read_buf):
                    raise exceptions.PipeNTError("Failed to write all data.")
        except pywintypes.error:
            self._handle_winerror()
        finally:
            # we disconnect the pipes at the end of the loop instead of at __exit__,
            # else the stdin process will be stuck waiting for more data as the context
            # has not ended
            with suppress(pywintypes.error):
                self._disconnect_pipe(self._pipe_data_stdin)
                self._disconnect_pipe(self._pipe_data_stdout)

            logging.getLogger("console").debug(
                f"Pipe bridge for {self._pipe_data_stdout.pipe_name} to "
                f"{self._pipe_data_stdin.pipe_name} ended"
            )

    def _create_pipe(self, data: PipeNamedNT.PipeData) -> None:
        """Create named pipe."""
        try:
            pipe = win32pipe.CreateNamedPipe(  # pyright: ignore [reportUnknownMemberType]
                str(data.pipe_name),
                data.open_mode,
                data.pipe_mode,
                data.max_instances,
                data.out_buffer_size,
                data.in_buffer_size,
                data.time_out,
                data.security_attributes,
            )

            logging.getLogger("console").debug(f"Created pipe {data.pipe_name}")

            data.file_handle = pipe
        except pywintypes.error:
            self._handle_winerror()

    def _connect_pipe(self, data: PipeData) -> None:
        """Connect to named pipe.

        This is blocking and needs to run in it's own thread.
        """
        try:
            if data.file_handle is not None:
                logging.getLogger("console").debug(f"Connecting {data.pipe_name}")
                if win32pipe.ConnectNamedPipe(data.file_handle, None):  # pyright: ignore [reportUnknownMemberType]
                    self._handle_winerror()

                logging.getLogger("console").debug(f"Connected {data.pipe_name}")

        except pywintypes.error:
            self._handle_winerror()

    def _close_pipe(self, path: str, mode: int) -> None:
        """Close a pipe.

        Due to ConnectNamedPipe being blocking, we must force close the pipe by
        opening the pipe and closing it. This will cause a connection to the pipe
        and an instant disconnect.
        """
        logging.getLogger("console").debug(
            f"Closing pipe {self.in_path} -> {self.out_path}"
        )

        with suppress(pywintypes.error):
            file = win32file.CreateFile(
                path,
                mode,
                0,
                win32security.SECURITY_ATTRIBUTES(),
                win32file.OPEN_EXISTING,
                0,
                None,
            )

            win32file.CloseHandle(file.handle)

    @staticmethod
    def _disconnect_pipe(data: PipeData) -> None:
        """Disconnnect a pipe."""
        if data.file_handle is None:
            raise exceptions.PipeError("Pipe handle does not exist.")

        logging.getLogger("console").debug(f"Disconnecting {data.pipe_name}")
        win32pipe.DisconnectNamedPipe(data.file_handle)

    @staticmethod
    def _handle_winerror(ret: int | None = None) -> None:
        """Print windows related errors."""
        match ret if ret is not None else win32.GetLastError():
            case winerror.ERROR_BROKEN_PIPE:
                # The pipe has been ended.
                return

            case _ as e:
                raise exceptions.PipeNTError(
                    f"Windows Named Pipes failed with error {e}. "
                    "If this persists consider using --two-step as an alternative."
                )



from __future__ import annotations

import logging
import os
from ast import literal_eval
from contextlib import suppress
from pathlib import Path
from typing import TYPE_CHECKING

from tbc_video_export.common import exceptions
from tbc_video_export.common.enums import (
    ChromaDecoder,
    TBCType,
    VideoSystem,
)
from tbc_video_export.common.utils import ansi
from tbc_video_export.opts.opts import AudioTrackOpt, Opts

if TYPE_CHECKING:
    import argparse
    from typing import Any

    from tbc_video_export.program_state import ProgramState


def validate_opts(
    state: ProgramState, parser: argparse.ArgumentParser, opts: Opts
) -> None:
    """Validate any nonsensical opt combinations."""
    _validate_line_opts(parser, opts)
    _validate_video_system(state, parser, opts)
    _validate_ansi_support(opts)
    _validate_luma_only_opts(parser, opts)
    _validate_video_format(parser, opts)
    _validate_decoder_opts(state, opts)


def _validate_line_opts(parser: argparse.ArgumentParser, opts: Opts) -> None:
    # check custom field/frame line opts
    field_frame_opts = [
        "first_active_field_line",
        "last_active_field_line",
        "first_active_frame_line",
        "last_active_frame_line",
    ]

    if any(getattr(opts, x) is not None for x in field_frame_opts) and (
        opts.vbi or opts.full_vertical or opts.letterbox
    ):
        parser.error(
            "arguments [--vbi | --letterbox]: not allowed with arguments "
            "[--ffll | --lfll | --ffrl | --lfrl]"
        )


def _validate_video_system(
    state: ProgramState, parser: argparse.ArgumentParser, opts: Opts
) -> None:
    # check video system incompatible opts
    match state.video_system:
        case VideoSystem.PAL | VideoSystem.PAL_M:
            if opts.oftest:
                parser.error(
                    "arguments --oftest: not allowed when video-system is not ntsc"
                )

            if opts.ntsc_phase_comp is not None:
                parser.error(
                    "arguments --ntsc-phase-comp/--no-ntsc-phase-comp: not allowed "
                    "when video-system is not NTSC"
                )

            if opts.chroma_nr is not None:
                parser.error(
                    "arguments --chroma-nr: not allowed when video-system is not ntsc"
                )

        case VideoSystem.NTSC:
            if opts.simple_pal:
                parser.error(
                    "arguments --simple-pal: not allowed when --video-system is ntsc"
                )


def _validate_video_format(parser: argparse.ArgumentParser, opts: Opts) -> None:
    # require bitdepth if format set
    if opts.video_format is not None and opts.video_bitdepth is None:
        parser.error("setting a video format requires a bitdepth.\n")


def _validate_ansi_support(opts: Opts) -> None:
    # check if ansi is supported on Windows and disable progress if not
    if not ansi.has_ansi_support():
        if os.name == "nt":
            logging.getLogger("console").critical(
                "Windows Version < 10.0.14393 (Windows 10 Anniversary Update 1607) "
                "detected.\n"
                "This version of Windows is unable to support ANSI escape sequences.\n"
                "The software will run but be unable to display realtime progress.\n"
            )

        logging.getLogger("console").critical(
            "No support for ANSI escape sequences.\n"
            "The software will run but be unable to display realtime progress.\n"
        )

        opts.no_progress = True


def _validate_luma_only_opts(parser: argparse.ArgumentParser, opts: Opts) -> None:
    # check luma only redundant opts
    if (opts.luma_only or opts.luma_4fsc) and opts.chroma_decoder is not None:
        parser.error(
            "arguments --chroma-decoder: not allowed with --luma-only or "
            "--luma-4fsc (redundant)"
        )


def valiate_metadata_file_exists(value: str) -> Path:
    """Return metadata path if it exists."""
    if (path := Path(value)).is_file():
        return path.absolute()

    raise exceptions.FileIOError(f"Metadata file {value} not found.")


def validate_audio_track_opts(value: str) -> AudioTrackOpt:
    """Return AudioTrackOpt from string."""
    return AudioTrackOpt(Path(value).absolute())


def validate_audio_track_advanced_opts(value: str) -> AudioTrackOpt:
    """Validate input types for the audio track advanced object."""
    try:
        data: list[Any] = literal_eval(value)
        type_check: set[bool] = set()

        if not data:
            raise exceptions.FileIOError(
                "File path is required for ffmpeg track, see --help for examples."
            )

        # ensure input are correct types
        with suppress(IndexError):
            type_check.add(isinstance(data[0], str))
            type_check.add(isinstance(data[1], str | None))
            type_check.add(isinstance(data[2], str | None))
            type_check.add(isinstance(data[3], str | int | None))
            type_check.add(isinstance(data[4], str | None))
            type_check.add(isinstance(data[5], int | None))
            type_check.add(isinstance(data[6], str | None))
            type_check.add(isinstance(data[7], int | float | None))

        if False in type_check:
            raise SyntaxError

        # clone input and change file name to absolute path
        opts = data.copy()
        opts[0] = Path(data[0]).absolute()

        return AudioTrackOpt(*opts)
    except (SyntaxError, AttributeError) as e:
        raise exceptions.InvalidOptsError(
            "Invalid FFmpeg audio track opts, check --help for examples."
        ) from e


def validate_black_levels_opts(value: str) -> tuple[int, int, int] | None:
    """Validate black level opts.

    If a single value is provded we return it three times, if 3 values are provides we
    return all 3.
    """
    try:
        if values := value.split(","):
            if len(values) == 1:
                return (int(values[0]), int(values[0]), int(values[0]))

            if len(values) == 3:
                return (int(values[0]), int(values[1]), int(values[2]))
    except ValueError as e:
        raise exceptions.InvalidOptsError(
            "Invalid black level opts, check --help for examples."
        ) from e

    raise exceptions.InvalidOptsError(
        "Invalid black levels, check --help for examples."
    )


def _validate_decoder_opts(state: ProgramState, opts: Opts) -> None:
    """Validate chroma-decoder opts."""
    # this is only available on cvbs/ld or when --chroma-decoder-luma is used
    decoder = (
        state.decoder_chroma
        if TBCType.COMBINED in state.tbc_types
        else state.decoder_luma
    )

    if opts.luma_nr is not None and decoder is ChromaDecoder.MONO:
        raise exceptions.InvalidOptsError(
            "--luma-nr is not implemented with the mono decoder."
        )

    if opts.simple_pal and state.decoder_chroma not in [
        ChromaDecoder.TRANSFORM2D,
        ChromaDecoder.TRANSFORM3D,
    ]:
        raise exceptions.InvalidOptsError(
            "--simple-pal is only implemented with Transform2D/Transform3D."
        )



from __future__ import annotations

import json
from functools import cached_property
from pathlib import Path
from typing import Any

from tbc_video_export.common import exceptions
from tbc_video_export.common.enums import VideoSystem


class TBCJsonHelper:
    """Handles parsing the TBC json."""

    def __init__(self, file_name: Path | None, json_data: Any = None) -> None:
        if file_name is not None:
            self.file_name = file_name

            try:
                with Path.open(file_name, mode="r", encoding="utf-8") as json_file:
                    self._json_data = json.load(json_file)
            except FileNotFoundError as e:
                raise exceptions.TBCError(f"TBC json not found ({file_name}).") from e
            except PermissionError as e:
                raise exceptions.TBCError(
                    f"Permission denied opening TBC json ({file_name})."
                ) from e
            except json.JSONDecodeError as e:
                raise exceptions.TBCError(
                    f"Unable to parse TBC json ({file_name})."
                ) from e
        else:
            self._json_data = json.loads(json_data)

    @property
    def file_name(self) -> Path:
        """Return tbc json file name."""
        return self._file_name

    @file_name.setter
    def file_name(self, file_name: str | Path) -> None:
        self._file_name = Path(file_name)

    @cached_property
    def is_widescreen(self) -> bool:
        """Returns whether the json TBC flags widescreen."""
        return (
            "isWidescreen" in self._json_data["videoParameters"]
            and self._json_data["videoParameters"]["isWidescreen"]
        )

    @cached_property
    def video_system(self) -> VideoSystem:
        """Return VideoSystem from TBC json."""
        if "system" in self._json_data["videoParameters"]:
            system = self._json_data["videoParameters"]["system"]

            # search for PAL* or NTSC* in videoParameters.system
            # isSourcePal and isSourceNtsc sometimes used, but not
            # sure if it's worth checking for
            match system.replace("-", "_").lower():
                case VideoSystem.PAL.value:
                    return VideoSystem.PAL

                case VideoSystem.PAL_M.value:
                    return VideoSystem.PAL_M

                case VideoSystem.NTSC.value:
                    return VideoSystem.NTSC

                case _:
                    raise exceptions.TBCError(f"System unsupported ({system}).")

        raise exceptions.TBCError("Unable to read video system from TBC json.")

    @cached_property
    def field_count(self) -> int:
        """Get total # of fields in TBC."""
        return len(self._json_data["fields"])

    @cached_property
    def frame_count(self) -> int:
        """Get total # of frames in TBC."""
        return int(self.field_count / 2)

    @cached_property
    def timecode(self) -> str:
        """Attempt to read a VITC timecode for the first frame.

        Return starting timecode if no VITC data found.
        """
        if (
            not self._json_data["fields"]
            or "vitc" not in self._json_data["fields"][0]
            or "vitcData" not in self._json_data["fields"][0]["vitc"]
        ):
            return "00:00:00:00"

        is_valid = True
        is_30_frame = self.video_system is not VideoSystem.PAL
        vitc_data = self._json_data["fields"][0]["vitc"]["vitcData"]

        def decode_bcd(tens: int, units: int) -> int:
            nonlocal is_valid

            if tens > 9:
                is_valid = False
                tens = 9

            if units > 9:
                is_valid = False
                units = 9

            return (tens * 10) + units

        hour = decode_bcd(vitc_data[7] & 0x03, vitc_data[6] & 0x0F)
        minute = decode_bcd(vitc_data[5] & 0x07, vitc_data[4] & 0x0F)
        second = decode_bcd(vitc_data[3] & 0x07, vitc_data[2] & 0x0F)
        frame = decode_bcd(vitc_data[1] & 0x03, vitc_data[0] & 0x0F)

        invalid_time = hour > 23 or minute > 59 or second > 59

        if (
            invalid_time
            or (is_30_frame and frame > 29)
            or (not is_30_frame and frame > 24)
        ):
            is_valid = False

        is_drop_frame = (vitc_data[1] & 0x04) != 0 if is_30_frame else False

        if not is_valid:
            return "00:00:00:00"

        sep = ";" if is_drop_frame else ":"

        return f"{hour:02d}:{minute:02d}:{second:02d}{sep}{frame:02d}"



from __future__ import annotations

from tbc_video_export.process.parser.export_state import ExportState
from tbc_video_export.process.parser.parser import Parser
from tbc_video_export.process.parser.parser_factory import ParserFactory

__all__ = ["ExportState", "Parser", "ParserFactory"]



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Generic

from tbc_video_export.process.wrapper.pipe import (
    PipeInputGeneric,
    PipeOutputGeneric,
)

if TYPE_CHECKING:
    from tbc_video_export.common.enums import ExportMode, TBCType


@dataclass
class WrapperConfig(Generic[PipeInputGeneric, PipeOutputGeneric]):
    """Wrapper config class."""

    export_mode: ExportMode
    tbc_type: TBCType
    input_pipes: PipeInputGeneric
    output_pipes: PipeOutputGeneric



from __future__ import annotations

from typing import TYPE_CHECKING

from tbc_video_export.common import exceptions
from tbc_video_export.common.enums import (
    HardwareAccelType,
    VideoSystem,
)
from tbc_video_export.common.utils import FlatList, ansi

if TYPE_CHECKING:
    from tbc_video_export.config.json import (
        JsonProfile,
        JsonSubProfile,
        JsonSubProfileAudio,
        JsonSubProfileFilter,
        JsonSubProfileVideo,
    )


class Profile:
    """Holds profile data."""

    def __init__(
        self,
        profile: JsonProfile,
        video_profile: ProfileVideo,
        audio_profile: ProfileAudio | None,
        filter_profiles: list[ProfileFilter],
    ) -> None:
        self._profile = profile
        self.video_profile = video_profile
        self.audio_profile = audio_profile
        self._filter_profiles = filter_profiles

    @property
    def name(self) -> str:
        """Return profile name."""
        return self._profile["name"]

    @property
    def include_vbi(self) -> bool:
        """Returns True if the profile contains include_vbi as True."""
        return self._profile["include_vbi"] if "include_vbi" in self._profile else False

    @property
    def is_default(self) -> bool:
        """Returns True if the profile is flagged as default."""
        return self._profile["default"] if "default" in self._profile else False

    @property
    def filter_profiles(self) -> list[ProfileFilter]:
        """Return filter profiles."""
        return self._filter_profiles

    @filter_profiles.setter
    def filter_profiles(self, filter_profiles: list[ProfileFilter]) -> None:
        """Set filter profiles."""
        self._filter_profiles = filter_profiles

    def __str__(self) -> str:  # noqa: D105
        data = f"--{self.name} {'(default)' if self.is_default else ''}\n"

        data += str(self.video_profile)

        if self.audio_profile is not None:
            data += str(self.audio_profile)

        if self.include_vbi:
            data += f"  {ansi.dim('Include VBI')}\t{self.include_vbi}\n"

        return data


class SubProfile:
    """Abstract class for subprofiles."""

    def __init__(self, profile: JsonSubProfile):
        self._profile = profile

    @property
    def name(self) -> str:
        """Return profile name."""
        return self._profile["name"]

    @property
    def description(self) -> str:
        """Return the profile description."""
        return self._profile["description"]


class ProfileVideo(SubProfile):
    """Holds FFmpeg video profile."""

    def __init__(self, parent: JsonProfile, profile: JsonSubProfileVideo) -> None:
        super().__init__(profile)
        self._parent = parent
        self._profile = profile
        self._video_format = self._profile["video_format"]

    @property
    def container(self) -> str:
        """Return the output file container."""
        return self._profile["container"]

    @property
    def output_format(self) -> str | None:
        """Return the output format."""
        return (
            self._profile["output_format"] if "output_format" in self._profile else None
        )

    @property
    def codec(self) -> str:
        """Return the video codec."""
        return self._profile["codec"]

    @property
    def opts(self) -> FlatList | None:
        """Return the video opts if they exist."""
        return FlatList(self._profile["opts"]) if "opts" in self._profile else None

    @property
    def video_format(self) -> str:
        """Return the video format."""
        return self._video_format

    @video_format.setter
    def video_format(self, video_format: str) -> None:
        """Set video format."""
        self._video_format = video_format

    @property
    def filter_profiles_additions(self) -> list[str]:
        """Return the additional filters if they exists."""
        return (
            self._profile["filter_profiles_additions"]
            if "filter_profiles_additions" in self._profile
            else []
        )

    @property
    def filter_profiles(self) -> list[str]:
        """Return the filters to override parent filters if they exists."""
        return (
            self._profile["filter_profiles_override"]
            if "filter_profiles_override" in self._profile
            else self._parent["filter_profiles"]
            if "filter_profiles" in self._parent
            else []
        )

    @property
    def hardware_accel(self) -> HardwareAccelType | None:
        """Return the hardware accel opt."""
        try:
            return HardwareAccelType(self._profile["hardware_accel"])
        except (KeyError, ValueError):
            return None

    @property
    def video_system(self) -> VideoSystem | None:
        """Return the video system filter."""
        if "video_system" in self._profile:
            try:
                return VideoSystem(self._profile["video_system"])
            except (KeyError, ValueError) as e:
                raise exceptions.InvalidProfileError(
                    f"Video profile {self._profile['name']} contains unknown "
                    f"video_system."
                ) from e

        return None

    def __str__(self) -> str:  # noqa: D105
        data = (
            f"  --{ansi.bold(self.hardware_accel.value)} "
            if self.hardware_accel is not None
            else ""
        )

        if data == "  ":
            data += "default"

        data += "\n"
        data += f"    {ansi.dim('Description')}\t{self.description} [{self.name}]\n"
        data += f"    {ansi.dim('Video Codec')}\t{self.codec}\n"

        if self.opts is not None:
            data += f"    {ansi.dim('Video Opts')}\t{self.opts}\n"

        data += f"    {ansi.dim('Format')}\t{self.video_format}\n"
        data += f"    {ansi.dim('Container')}\t{self.container}"

        if self.output_format is not None:
            data += f" ({self.output_format})"

        data += "\n"

        if self.filter_profiles_additions or self.filter_profiles:
            filters = self.filter_profiles + self.filter_profiles_additions
            data += f"    {ansi.dim('Filters')}\t"
            data += f"{', '.join(filters) if filters else 'None'}\n"

        if self.video_system is not None:
            data += f"    {ansi.dim('System')}\t{self.video_system}\n"

        # data += "\n"

        return data


class ProfileAudio(SubProfile):
    """Holds FFmpeg audio profile."""

    def __init__(self, profile: JsonSubProfileAudio) -> None:
        super().__init__(profile)
        self._profile = profile

    @property
    def codec(self) -> str:
        """Return the audio codec."""
        return self._profile["codec"]

    @property
    def opts(self) -> FlatList:
        """Return the audio opts if they exist."""
        return (
            FlatList(self._profile["opts"]) if "opts" in self._profile else FlatList()
        )

    def __str__(self) -> str:  # noqa: D105
        data = f"  {ansi.dim('Audio Codec:')}\t{self.codec}\n"
        if self.opts:
            data += f"  {ansi.dim('Audio Opts')}\t{self.opts}\n"

        return data


class ProfileFilter(SubProfile):
    """Holds FFmpeg filter profile."""

    def __init__(self, profile: JsonSubProfileFilter) -> None:
        super().__init__(profile)
        self._profile = profile

    @property
    def video_filter(self) -> str | None:
        """Return the video filter."""
        return self._profile.get("video_filter", None)

    @property
    def other_filter(self) -> str | None:
        """Return the other filters.

        These go after the main video filter and can contain any other type of filter.
        """
        return self._profile.get("other_filter", None)



from __future__ import annotations

import argparse
from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pathlib import Path

    from tbc_video_export.common.enums import (
        ChromaDecoder,
        FieldOrder,
        HardwareAccelType,
        VideoFormatType,
        VideoSystem,
    )


class Opts(argparse.Namespace):
    """Program opts namespace.

    We define available options and their types to keep the linter happy.
    """

    # general
    input_file: str
    output_file: str | None
    threads: int
    appimage: str
    two_step: bool
    async_nt_pipes: bool
    video_system: VideoSystem | None
    input_tbc_json: str | None
    overwrite: bool
    dry_run: bool
    dump_default_config: bool

    # verbosity
    quiet: bool
    debug: bool
    no_debug_log: bool
    no_progress: bool
    show_process_output: bool
    log_process_output: bool

    # decoder
    start: int | None
    length: int | None
    reverse: bool
    output_padding: int
    vbi: bool
    full_vertical: bool
    letterbox: bool
    first_active_field_line: int | None
    last_active_field_line: int | None
    first_active_frame_line: int | None
    last_active_frame_line: int | None
    chroma_decoder: ChromaDecoder | None
    chroma_decoder_luma: ChromaDecoder | None
    chroma_gain: float | None
    chroma_phase: float | None
    luma_nr: float | None
    transform_threshold: float | None
    transform_thresholds: str | None

    # decoder (ntsc)
    ntsc_phase_comp: bool | None
    oftest: bool
    chroma_nr: float | None

    # decoder (pal)
    simple_pal: bool

    # dropout-correct
    no_dropout_correct: bool

    # luma
    luma_only: bool
    luma_4fsc: bool

    # process-vbi
    process_vbi: bool
    process_vbi_keep_going: bool

    # process-efm
    process_efm: bool
    process_efm_keep_going: bool
    process_efm_dts: bool

    # export-metadata
    export_metadata: bool
    export_metadata_keep_going: bool

    # ffmpeg
    audio_track: list[AudioTrackOpt]
    metadata: list[list[str]]
    metadata_file: list[Path]
    field_order: FieldOrder
    force_anamorphic: bool
    force_black_level: tuple[int, int, int] | None
    hwaccel_device: str | None
    no_attach_json: bool
    thread_queue_size: int
    checksum: bool

    # profile
    profile: str
    profile_luma: str

    # profile overrides
    profile_container: str | None
    profile_additional_filters: list[str]
    append_video_filter: str | None
    append_other_filter: str | None

    hwaccel_type: HardwareAccelType | None
    video_format: VideoFormatType | None
    video_bitdepth: int | None
    audio_profile: str | None

    def convert_opt(
        self, program_opt_name: str, target_opt_name: str
    ) -> str | tuple[str, str] | None:
        """Convert a program opt to a subprocess opt (list[str])."""
        if (value := getattr(self, program_opt_name)) is not None:
            if isinstance(value, bool):
                if value:
                    return target_opt_name
            else:
                return (target_opt_name, str(value))
        return None


@dataclass
class AudioTrackOpt:
    """Container class for audio track data."""

    file_name: Path
    title: str | None = None
    language: str | None = None

    rate: str | int | None = None
    sample_format: str | None = None
    channels: int | None = None
    layout: str | None = None
    offset: int | float | None = None


@dataclass
class MetaDataOpt:
    """Container class for meta data."""

    key: str
    value: str | int | float



from __future__ import annotations

from typing import TYPE_CHECKING

from tbc_video_export.common.enums import ProcessName
from tbc_video_export.process.parser.parser_ffmpeg import ParserFFmpeg
from tbc_video_export.process.parser.parser_ld_chroma_decoder import (
    ParserLDChromaDecoder,
)
from tbc_video_export.process.parser.parser_ld_dropout_correct import (
    ParserLDDropoutCorrect,
)
from tbc_video_export.process.parser.parser_ld_export_metadata import (
    ParserLDExportMetadata,
)
from tbc_video_export.process.parser.parser_ld_process_efm import ParserLDProcessEFM
from tbc_video_export.process.parser.parser_ld_process_vbi import ParserLDProcessVBI

if TYPE_CHECKING:
    from tbc_video_export.process.parser.parser import Parser


class ParserFactory:
    """Factory class for process parsing."""

    @classmethod
    def create(cls, process_name: ProcessName) -> Parser:
        """Create an output parser based on process name."""
        match process_name:
            case ProcessName.FFMPEG:
                return ParserFFmpeg(process_name)

            case ProcessName.LD_CHROMA_DECODER:
                return ParserLDChromaDecoder(process_name)

            case ProcessName.LD_DROPOUT_CORRECT:
                return ParserLDDropoutCorrect(process_name)

            case ProcessName.LD_PROCESS_VBI:
                return ParserLDProcessVBI(process_name)

            case ProcessName.LD_EXPORT_METADATA:
                return ParserLDExportMetadata(process_name)

            case ProcessName.LD_PROCESS_EFM:
                return ParserLDProcessEFM(process_name)

            case ProcessName.NONE:
                raise NotImplementedError(
                    f"Parser for process {process_name} not implemented."
                )



from __future__ import annotations

import asyncio
from functools import cached_property
from typing import TYPE_CHECKING

from tbc_video_export.common import exceptions
from tbc_video_export.common.enums import PipeType, ProcessName
from tbc_video_export.common.utils import FlatList
from tbc_video_export.process.wrapper.wrapper import Wrapper

if TYPE_CHECKING:
    from pathlib import Path

    from tbc_video_export.process.wrapper.wrapper_config import WrapperConfig
    from tbc_video_export.program_state import ProgramState


class WrapperLDProcessEFM(Wrapper):
    """Wrapper for ld-process-efm."""

    def __init__(self, state: ProgramState, config: WrapperConfig[None, None]) -> None:
        super().__init__(state, config)
        self._config = config

    def post_fn(self) -> None:  # noqa: D102
        pass

    @property
    def command(self) -> FlatList:  # noqa: D102
        return FlatList(
            (
                self.binary,
                self._state.opts.convert_opt("process_efm_dts", "--dts"),
                self._get_efm_file(),
                self._get_output_file(),
            ),
        )

    def _get_efm_file(self) -> Path:
        if self._state.file_helper.efm_file is None:
            raise exceptions.FileIOError(
                "Could not find EFM file. Try without --process-efm."
            )

        return self._state.file_helper.efm_file

    def _get_output_file(self) -> Path:
        return (
            self._state.file_helper.get_output_file_from_ext("digital.pcm")
            if not self._state.opts.process_efm_dts
            else self._state.file_helper.get_output_file_from_ext("dts")
        )

    @cached_property
    def process_name(self) -> ProcessName:  # noqa: D102
        return ProcessName.LD_PROCESS_EFM

    @cached_property
    def supported_pipe_types(self) -> PipeType:  # noqa: D102
        return PipeType.NONE

    @cached_property
    def stdin(self) -> int | None:  # noqa: D102
        return None

    @cached_property
    def stdout(self) -> int | None:  # noqa: D102
        return asyncio.subprocess.DEVNULL

    @cached_property
    def stderr(self) -> int | None:  # noqa: D102
        return asyncio.subprocess.PIPE

    @cached_property
    def log_output(self) -> bool:  # noqa: D102
        return True

    @cached_property
    def log_stdout(self) -> bool:  # noqa: D102
        return False

    @cached_property
    def env(self) -> dict[str, str] | None:  # noqa: D102
        return None

    @cached_property
    def ignore_error(self) -> bool:  # noqa: D102
        return False

    @cached_property
    def stop_on_last_alive(self) -> bool:  # noqa: D102
        return False



from __future__ import annotations

import logging
from dataclasses import dataclass

from tbc_video_export.common.enums import ProcessStatus


@dataclass
class ProcessState:
    """Process state class."""

    status = ProcessStatus.NONE
    returncode = -1
    status_index = -1

    def set_has_run(self) -> None:
        """Set the has run flag."""
        self.status |= ProcessStatus.HAS_RUN

    def set_running(self) -> None:
        """Set the running flag and removes the stopped flag."""
        self.status |= ProcessStatus.RUNNING
        self.status &= ~ProcessStatus.STOPPED

    def set_stopped(self) -> None:
        """Set the stopped flag and removes the running flag."""
        self.status |= ProcessStatus.STOPPED
        self.status &= ~ProcessStatus.RUNNING
        logging.getLogger("console").debug("Setting STOPPED")

    def set_success(self) -> None:
        """Set the success flag and removes the error flag."""
        self.status |= ProcessStatus.SUCCESS
        self.status &= ~ProcessStatus.ERROR
        logging.getLogger("console").debug("Setting SUCCESS")

    def set_error(self) -> None:
        """Set the error flag and removes the success flag."""
        self.status |= ProcessStatus.ERROR
        self.status &= ~ProcessStatus.SUCCESS
        logging.getLogger("console").debug("Setting ERROR")

    @property
    def has_run(self) -> bool:
        """Returns True if the process has run."""
        return ProcessStatus.HAS_RUN in self.status

    @property
    def is_successful(self) -> bool:
        """Returns True if the process is successful."""
        return ProcessStatus.SUCCESS in self.status

    @property
    def is_errored(self) -> bool:
        """Returns True if the process has errored.."""
        return ProcessStatus.ERROR in self.status

    @property
    def is_running(self) -> bool:
        """Returns True if the process is running."""
        return ProcessStatus.RUNNING in self.status

    @property
    def is_stopped(self) -> bool:
        """Returns True if the process is stopped."""
        return ProcessStatus.STOPPED in self.status



from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING

from tbc_video_export.common.enums import PipeType
from tbc_video_export.process.wrapper.pipe.pipe import Pipe

if TYPE_CHECKING:
    from pathlib import Path
    from types import TracebackType


class PipeDummy(Pipe):
    """Dummy pipe, does nothing. Used for dry-runs."""

    def __init__(self, stdin_str: Path | str, stdout_str: Path | str) -> None:
        self._stdin_str = stdin_str
        self._stdout_str = stdout_str

    async def __aenter__(self) -> Pipe:
        """Enter dummy pipe context."""
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None | bool:
        """Exit dummy pipe context."""

    @cached_property
    def pipe_type(self) -> PipeType:  # noqa: D102
        return PipeType.NULL

    @cached_property
    def in_path(self) -> Path | str:  # noqa: D102
        return self._stdin_str

    @cached_property
    def out_path(self) -> Path | str:  # noqa: D102
        return self._stdout_str

    @property
    def in_handle(self) -> int | None:  # noqa: D102
        return None

    @property
    def out_handle(self) -> int | None:  # noqa: D102
        return None

    def close(self) -> None:
        """Does nothing."""



from __future__ import annotations

from typing import TYPE_CHECKING

from tbc_video_export.common.enums import FieldOrder
from tbc_video_export.opts import opt_types, opt_validators

if TYPE_CHECKING:
    import argparse


def add_ffmpeg_opts(parent: argparse.ArgumentParser) -> None:
    """Add FFmpeg opts to the parent arg parser."""
    # ffmpeg arguments
    ffmpeg_opts = parent.add_argument_group("ffmpeg")

    ffmpeg_opts.add_argument(
        "--audio-track",
        dest="audio_track",
        action="append",
        default=[],
        type=opt_validators.validate_audio_track_opts,
        metavar="file_name",
        help="Audio track to mux.\nYou can use this option multiple times.\n\n",
    )

    ffmpeg_opts.add_argument(
        "--audio-track-advanced",
        dest="audio_track",
        action="append",
        default=[],
        type=opt_validators.validate_audio_track_advanced_opts,
        metavar=(
            "["
            "file_name, "
            "title, "
            "language, "
            "rate, "
            "sample_format, "
            "channels, "
            "channel_layout, "
            "offset"
            "]"
        ),
        help="Audio track to mux (advanced).\n"
        "You can use this option multiple times.\n"
        "Only file_name is required. None must be used when skipping an argument.\n\n"
        "Examples:\n"
        '\'["/path/to/file.flac", "HiFi", "eng", 192000]\'\n'
        '\'["/path/to/file.flac", "Linear", "eng", None, None, None, None, 0.15]\'\n'
        '\'["/path/to/file.pcm", "Analog Audio", None, 44100, "s16le", 2]\'\n'
        '\'["/path/to/file.dts", "PCM Surround", "eng", 44100, "s16le", 6, "5.1"]\'\n'
        "\n\n",
    )

    ffmpeg_opts.add_argument(
        "--metadata",
        nargs=2,
        default=[],
        action="append",
        metavar=("key", "value"),
        help="Add metadata to output the file.\n"
        "You can use this option multiple times.\n"
        "Example:\n\n"
        "--metadata Title foo --metadata Year 2024"
        "\n\n",
    )

    ffmpeg_opts.add_argument(
        "--metadata-file",
        type=opt_validators.valiate_metadata_file_exists,
        default=[],
        action="append",
        metavar="filename",
        help="Add metadata to the output of the file using ffmetadata files.\n"
        "You can use this option multiple times.\n\n"
        "See https://ffmpeg.org/ffmpeg-formats.html#Metadata-1 for details.\n\n"
        "Note: When using --export-metadata the generated ffmetadata file is also "
        "used.\n"
        "Files defined here are used before any generated metadata and take "
        "priority.\n"
        "\n\n",
    )

    ffmpeg_opts.add_argument(
        "--field-order",
        type=opt_types.TypeFieldOrder(parent),
        choices=list(FieldOrder),
        default=FieldOrder.TFF,
        metavar="order",
        help="Set a field order. (default: tff)\n"
        "Available formats:\n\n  "
        + "\n  ".join(f"{e.name.lower()!s:<5s} {e.value}" for e in FieldOrder)
        + "\n\n",
    )

    ffmpeg_opts.add_argument(
        "--force-anamorphic",
        action="store_true",
        default=False,
        help="Force widescreen aspect ratio.\n\n",
    )

    ffmpeg_opts.add_argument(
        "--force-black-level",
        type=opt_validators.validate_black_levels_opts,
        metavar="R[,G,B]",
        default=None,
        help="Force black levels using the colorlevels filter.\n"
        "Use a comma-separated list of numbers to provide values for colorlevels.\n"
        "If a single number is provided, it is used for all 3."
        "\n\n",
    )

    ffmpeg_opts.add_argument(
        "--hwaccel-device",
        type=str,
        help="Hardware acceleration device.\n"
        "Specify a hardware device to use when a supported video profile is selected."
        "\n\n",
    )

    ffmpeg_opts.add_argument(
        "--no-attach-json",
        action="store_true",
        default=False,
        help="Disable embedding the TBC json in the video file. (default: no).\n\n",
    )

    ffmpeg_opts.add_argument(
        "--thread-queue-size",
        type=int,
        default=1024,
        metavar="int",
        help="Set the thread queue size for FFmpeg. (default: 1024)\n"
        "Reduce this if you are having out-of-memory issues as a higher value\n"
        "will allow the decoder to consume more memory."
        "\n\n",
    )

    ffmpeg_opts.add_argument(
        "--checksum",
        action="store_true",
        default=False,
        help="Enable SHA256 checksumming on the output streams. (default: no)\n"
        "This will create a .sha256 file next to your output file.\n"
        "This may reduce export FPS slightly. FFmpeg must be used to verify checksums."
        "\n\n",
    )



from __future__ import annotations

import logging
import os
from contextlib import suppress
from functools import cached_property
from pathlib import Path
from tempfile import mkdtemp
from typing import TYPE_CHECKING

from tbc_video_export.common import consts, exceptions
from tbc_video_export.common.enums import PipeType
from tbc_video_export.process.wrapper.pipe.pipe import Pipe

if TYPE_CHECKING:
    from types import TracebackType

    from tbc_video_export.common.enums import ProcessName, TBCType


assert os.name == "posix"  # keep the type checker happy


class PipeNamedPosix(Pipe):
    """Named pipe helper for POSIX systems."""

    # use a class variable as we only want a single temp dir created
    tmp_dir: str = ""

    def __init__(self, process_name: ProcessName, tbc_type: TBCType) -> None:
        super().__init__(process_name, tbc_type)

        if not PipeNamedPosix.tmp_dir:
            try:
                PipeNamedPosix.tmp_dir = mkdtemp(
                    prefix=f"{consts.APPLICATION_NAME}-", suffix=""
                )
            except PermissionError as e:
                raise exceptions.PipeError(
                    "Unable to create pipes due to permissions."
                ) from e

    async def __aenter__(self) -> Pipe:
        """Enter posix pipe context."""
        try:
            logging.getLogger("console").debug(f"Creating named pipe {self.in_path}")
            os.mkfifo(self.in_path)

            return self
        except FileNotFoundError as e:
            raise exceptions.PipeError(
                "Unable to create pipes due to permissions."
            ) from e
        except PermissionError as e:
            raise exceptions.PipeError(
                "Unable to create pipes due to permissions."
            ) from e

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None | bool:
        """Exit posix pipe context."""
        PipeNamedPosix.tmp_dir = ""
        self.close()

    @cached_property
    def pipe_type(self) -> PipeType:  # noqa: D102
        return PipeType.NAMED_POSIX

    @cached_property
    def in_path(self) -> Path | str:  # noqa: D102
        return Path(PipeNamedPosix.tmp_dir).joinpath(
            f"{self._process_name}-{self._tbc_type}"
        )

    @cached_property
    def out_path(self) -> Path | str:
        """Get pipe stdout string.

        POSIX systems use the same path for in/out.
        """
        return self.in_path

    @property
    def in_handle(self) -> int | None:  # noqa: D102
        return None

    @property
    def out_handle(self) -> int | None:  # noqa: D102
        return None

    def close(self) -> None:
        """Close pipe."""
        logging.getLogger("console").debug(f"Closing pipe {self.in_path}")
        # we suppress FileNotFoundError so other files can be cleaned up
        with suppress(FileNotFoundError, PermissionError):
            Path.unlink(Path(self.in_path))

            # other pipes may exist in the directory, only remove if empty
            if len(os.listdir(PipeNamedPosix.tmp_dir)) == 0:
                Path.rmdir(Path(PipeNamedPosix.tmp_dir))



from __future__ import annotations

import re
from contextlib import suppress
from typing import TYPE_CHECKING

from tbc_video_export.process.parser.export_state import ExportStateSnapshot
from tbc_video_export.process.parser.parser import Parser

if TYPE_CHECKING:
    from tbc_video_export.common.enums import ProcessName


class ParserLDDropoutCorrect(Parser):
    """Parser for ld-dropout-correct process."""

    def __init__(self, process_name: ProcessName) -> None:
        super().__init__(process_name)

        self.tracked_value_name = "frame"

    def parse_line(self, line: str) -> ExportStateSnapshot:  # noqa: D102
        state = ExportStateSnapshot()

        patterns = [
            r"Info: Using .* threads to process (.*?) frames",
            r"Info: Processed and written frame (.*?)$",
            (
                r"Info: Dropout correction complete - (.*?) "
                r"frames in .* seconds \( (.*?) FPS \)$"
            ),
            r"Info:   Total concealments: (.*?)$",
            r"Critical: .*",
        ]

        with suppress(ValueError):
            if (reg := re.match("|".join(patterns), line)) is not None:
                # matches but no groups, an error
                if all(v is None for v in reg.groups()):
                    self.error_count += 1
                    state.message = self._create_log_line(line)
                elif (frame_count := reg.group(1)) is not None:
                    self.tracked_value_total = int(frame_count)
                elif (current_frame := reg.group(2)) is not None:
                    self.tracked_value = int(current_frame)
                elif (total_frames := reg.group(3)) is not None and (
                    (end_fps := reg.group(4)) is not None
                ):
                    self.tracked_value = int(total_frames)
                    self.fps = float(end_fps)
                elif (concealments := reg.group(5)) is not None:
                    state.concealments = int(concealments)

        return state

    @property
    def hide_tbc_type(self) -> bool:  # noqa: D102
        return False



from __future__ import annotations

import asyncio
from functools import cached_property
from typing import TYPE_CHECKING

from tbc_video_export.common import exceptions
from tbc_video_export.common.enums import (
    ChromaDecoder,
    PipeType,
    ProcessName,
    TBCType,
    VideoSystem,
    VideoSystemLines,
)
from tbc_video_export.common.utils import FlatList
from tbc_video_export.process.wrapper.wrapper import Wrapper

if TYPE_CHECKING:
    from tbc_video_export.process.wrapper.pipe import Pipe
    from tbc_video_export.process.wrapper.wrapper_config import WrapperConfig
    from tbc_video_export.program_state import ProgramState


class WrapperLDChromaDecoder(Wrapper):
    """Wrapper for ld-tools-decoder."""

    def __init__(self, state: ProgramState, config: WrapperConfig[Pipe, Pipe]) -> None:
        super().__init__(state, config)
        self._config = config

    def post_fn(self) -> None:  # noqa: D102
        pass

    @property
    def command(self) -> FlatList:  # noqa: D102
        return FlatList(
            (
                self.binary,
                self._get_gain_nr_opts(),
                "-p",
                "y4m",
                self._get_decoder_opts(),
                self._get_active_line_opts(),
                self._get_misc_opts(),
                "--input-json",
                self._state.file_helper.tbc_json.file_name,
                self._config.input_pipes.in_path,
                self._config.output_pipes.out_path,
            )
        )

    def _get_gain_nr_opts(self) -> FlatList:
        """Return ld-chroma-decoder opts."""
        gain_nr_opts = FlatList()

        if self._config.tbc_type in (TBCType.LUMA, TBCType.COMBINED):
            gain_nr_opts.append(self._state.opts.convert_opt("luma_nr", "--luma-nr"))

        if self._config.tbc_type in (TBCType.CHROMA, TBCType.COMBINED):
            gain_nr_opts.append(
                (
                    self._state.opts.convert_opt("chroma_gain", "--chroma-gain"),
                    self._state.opts.convert_opt("chroma_nr", "--chroma-nr"),
                    self._state.opts.convert_opt("chroma_phase", "--chroma-phase"),
                )
            )

        # set defaults for separated TBC
        if self._config.tbc_type is TBCType.LUMA:
            # chroma-gain not used by mono decoder, set anyway
            gain_nr_opts.append(("--chroma-gain", "0"))

        if self._config.tbc_type is TBCType.CHROMA:
            gain_nr_opts.append(("--luma-nr", "0"))

        return gain_nr_opts

    def _get_decoder_opts(self) -> FlatList:
        """Return decoder to use."""
        # validate decoders for video system
        decoder = (
            self._state.decoder_chroma
            if self.tbc_type is not TBCType.LUMA
            else self._state.decoder_luma
        )

        match self._state.video_system:
            case VideoSystem.PAL | VideoSystem.PAL_M:
                if decoder not in {
                    ChromaDecoder.MONO,
                    ChromaDecoder.PAL2D,
                    ChromaDecoder.TRANSFORM2D,
                    ChromaDecoder.TRANSFORM3D,
                }:
                    raise exceptions.InvalidChromaDecoderError(
                        f"{decoder} is not a valid decoder for "
                        f"{self._state.video_system}."
                    )

            case VideoSystem.NTSC:
                if decoder not in {
                    ChromaDecoder.MONO,
                    ChromaDecoder.NTSC1D,
                    ChromaDecoder.NTSC2D,
                    ChromaDecoder.NTSC3D,
                    ChromaDecoder.NTSC3DNOADAPT,
                }:
                    raise exceptions.InvalidChromaDecoderError(
                        f"{decoder} is not a valid decoder for "
                        f"{self._state.video_system}."
                    )

        return FlatList(
            (
                "-f",
                self._state.decoder_luma.value
                if self.tbc_type is TBCType.LUMA
                else self._state.decoder_chroma.value,
            )
        )

    def _get_active_line_opts(self) -> FlatList:
        """Return active line opts."""
        values: tuple[int, ...] | None = None

        if (
            self._state.opts.full_vertical
            or self._state.opts.vbi
            or self._state.profile.include_vbi
        ):
            match self._state.video_system:
                case VideoSystem.PAL:
                    values = VideoSystemLines.PAL_FULL_VERTICAL.value

                case VideoSystem.NTSC | VideoSystem.PAL_M:
                    values = VideoSystemLines.NTSC_FULL_VERTICAL.value

        if self._state.opts.letterbox:
            match self._state.video_system:
                case VideoSystem.PAL:
                    values = VideoSystemLines.PAL_LETTERBOX.value

                case VideoSystem.NTSC:
                    values = VideoSystemLines.NTSC_LETTERBOX.value

                case VideoSystem.PAL_M:
                    raise exceptions.SampleRequiredError(
                        f"{str(self._state.video_system).upper()} letterbox"
                    )

        if values is not None:
            return FlatList(
                (
                    "--ffll",
                    values[0],
                    "--lfll",
                    values[1],
                    "--ffrl",
                    values[2],
                    "--lfrl",
                    values[3],
                )
            )

        # use default/user values
        return FlatList(
            (
                self._state.opts.convert_opt("first_active_field_line", "--ffll"),
                self._state.opts.convert_opt("last_active_field_line", "--lfll"),
                self._state.opts.convert_opt("first_active_frame_line", "--ffrl"),
                self._state.opts.convert_opt("last_active_frame_line", "--lfrl"),
            ),
        )

    def _get_misc_opts(self) -> FlatList:
        """Return ld-chroma-decoder opts."""
        decoder_opts = FlatList()

        decoder_opts.append(
            (
                self._state.opts.convert_opt("start", "-s"),
                self._state.opts.convert_opt("length", "-l"),
                self._state.opts.convert_opt("reverse", "-r"),
                self._state.opts.convert_opt("threads", "-t"),
                self._state.opts.convert_opt("output_padding", "--pad"),
                self._state.opts.convert_opt("oftest", "-o"),
                self._state.opts.convert_opt("simple_pal", "--simple-pal"),
                self._state.opts.convert_opt(
                    "transform_threshold", "--transform-threshold"
                ),
                self._state.opts.convert_opt(
                    "transform_thresholds", "--transform-thresholds"
                ),
            ),
        )

        if self._state.video_system is VideoSystem.NTSC:
            # True unless ld detected
            add_phase_check = not self._state.file_helper.is_combined_ld

            # override if user has set
            if self._state.opts.ntsc_phase_comp is not None:
                add_phase_check = self._state.opts.ntsc_phase_comp

            if add_phase_check:
                decoder_opts.append("--ntsc-phase-comp")

        return decoder_opts

    @cached_property
    def process_name(self) -> ProcessName:  # noqa: D102
        return ProcessName.LD_CHROMA_DECODER

    @cached_property
    def supported_pipe_types(self) -> PipeType:  # noqa: D102
        return PipeType.NULL | PipeType.OS | PipeType.NAMED

    @cached_property
    def stdin(self) -> int | None:  # noqa: D102
        return self._config.input_pipes.in_handle

    @cached_property
    def stdout(self) -> int | None:  # noqa: D102
        return self._config.output_pipes.out_handle

    @cached_property
    def stderr(self) -> int | None:  # noqa: D102
        return asyncio.subprocess.PIPE

    @cached_property
    def log_output(self) -> bool:  # noqa: D102
        return True

    @cached_property
    def log_stdout(self) -> bool:  # noqa: D102
        return False

    @cached_property
    def env(self) -> dict[str, str] | None:  # noqa: D102
        return None

    @cached_property
    def ignore_error(self) -> bool:  # noqa: D102
        return False

    @cached_property
    def stop_on_last_alive(self) -> bool:  # noqa: D102
        return False



from __future__ import annotations

from abc import ABC, abstractmethod
from functools import cached_property
from typing import TYPE_CHECKING

from tbc_video_export.common import exceptions
from tbc_video_export.common.enums import FlagHelper, PipeType
from tbc_video_export.common.utils import FlatList
from tbc_video_export.process.wrapper.pipe import (
    Pipe,
    PipeInputGeneric,
    PipeOutputGeneric,
)

if TYPE_CHECKING:
    from tbc_video_export.common.enums import ProcessName, TBCType
    from tbc_video_export.process.wrapper.wrapper_config import WrapperConfig
    from tbc_video_export.program_state import ProgramState


class Wrapper(ABC):
    """Abstract Wrapper class."""

    def __init__(
        self,
        state: ProgramState,
        config: WrapperConfig[PipeInputGeneric, PipeOutputGeneric],
    ) -> None:
        self._state = state
        self._config = config

        # trigger any wrapper exceptions early
        _ = self.command
        self._check_pipes()

    @property
    def binary(self) -> FlatList:
        """Return wrapped binary name."""
        return FlatList(self._state.file_helper.get_tool(self.process_name))

    @cached_property
    def tbc_type(self) -> TBCType:
        """Return TBC type for the wrapper instance."""
        return self._config.tbc_type

    @cached_property
    def pipes(self) -> tuple[Pipe, ...]:
        """Return pipes for wrapper."""
        pipes: tuple[Pipe, ...] = ()
        in_pipe = self._config.input_pipes
        out_pipe = self._config.output_pipes

        if isinstance(in_pipe, tuple):
            pipes += in_pipe
        elif isinstance(in_pipe, Pipe):
            pipes = (*pipes, in_pipe)

        if isinstance(out_pipe, tuple):
            pipes += out_pipe
        elif isinstance(out_pipe, Pipe):
            pipes = (*pipes, out_pipe)

        return pipes

    @abstractmethod
    def post_fn(self) -> None:
        """Post-process function.

        This will run after the process has finished running.
        """

    @property
    @abstractmethod
    def command(self) -> FlatList:
        """Return wrapped process arguments."""

    @cached_property
    @abstractmethod
    def process_name(self) -> ProcessName:
        """Return process name."""

    @cached_property
    @abstractmethod
    def supported_pipe_types(self) -> PipeType:
        """Supported pipe types of the wrapped process."""

    @cached_property
    @abstractmethod
    def stdin(self) -> int | None:
        """Return process stdin."""

    @cached_property
    @abstractmethod
    def stdout(self) -> int | None:
        """Return process stdout."""

    @cached_property
    @abstractmethod
    def stderr(self) -> int | None:
        """Return process stderr."""

    @cached_property
    @abstractmethod
    def log_output(self) -> bool:
        """Log output of process to file."""

    @cached_property
    @abstractmethod
    def log_stdout(self) -> bool:
        """Log stdout instead of stderr."""

    @cached_property
    @abstractmethod
    def env(self) -> dict[str, str] | None:
        """List of env args to run the process with."""

    @cached_property
    @abstractmethod
    def ignore_error(self) -> bool:
        """Ignore errors from process."""

    @cached_property
    @abstractmethod
    def stop_on_last_alive(self) -> bool:
        """Stop process of this type if it"s the last remaining."""

    def _check_pipes(self) -> None:
        """Ensure pipes are valid for wrapper."""
        # ensure all pipes are supported
        if sum(
            1
            for pipe in self.pipes
            if pipe.pipe_type not in FlagHelper.get_flags(self.supported_pipe_types)
        ):
            pipes_str = ""

            for pipe in self.pipes:
                pipes_str += str(pipe)

            raise exceptions.PipeError(
                f"{self.process_name} unable to use pipe types.\n{pipes_str}"
            )



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from tbc_video_export.common import consts
from tbc_video_export.common.utils import ansi

if TYPE_CHECKING:
    import asyncio
    from pathlib import Path
    from typing import Any


def _print_exception(e: BaseException) -> None:
    """Print formatted exception with exception name and message."""
    if len(message := getattr(e, "message", str(e))) > 1:
        logging.getLogger("console").exception(
            ansi.error_color(f"{e.__class__.__name__}: {message}"),
            exc_info=False,
        )


def handle_exceptions(e: BaseException | None):
    """Handle exceptions for the application.

    Provides formatted and custom error messages for exceptions.
    """
    match e:
        case (
            TBCError()
            | TBCTypeError()
            | FileIOError()
            | InvalidOptsError()
            | PipeError()
            | PipeNTError()
            | ProcessError()
        ):
            _print_exception(e)
        case InvalidProfileError():
            _print_exception(e)
            logging.getLogger("console").critical(f"\nError parsing {e.config_path}.")
            logging.getLogger("console").critical(
                "If you have upgraded this config file may not be compatible.\n"
                "Try moving or deleting the file and trying again."
            )
        case SampleRequiredError():
            logging.getLogger("console").critical(
                f"Unable to export file due to unsupported options: {e}\n"
                f"Please provide a capture sample via GitHub or Discord for this to be "
                f"fixed in future releases.\n\n"
                f"{consts.PROJECT_URL_ISSUES}\n{consts.PROJECT_URL_DISCORD}"
            )

        case KeyboardInterrupt():
            logging.getLogger("console").critical("User has cancelled the export.")

        case None:
            pass

        case _:
            _print_exception(e)


def loop_exception_handler(loop: asyncio.AbstractEventLoop, context: dict[str, Any]):  # noqa: ARG001
    """Loop exception handler."""
    handle_exceptions(context.get("exception"))


class TBCError(Exception):
    """TBC errors."""


class TBCTypeError(Exception):
    """TBC type errors."""


class FileIOError(Exception):
    """General file i/o errors."""


class SampleRequiredError(Exception):
    """Unsupported option due to lack of sample error."""


class InvalidProfileError(Exception):
    """General profile errors."""

    def __init__(self, message: str, config_path: Path | None = None) -> None:
        super().__init__(message)
        self.config_path = config_path if config_path is not None else "[internal]"


class InvalidVideoProfileError(InvalidProfileError):
    """Video profile errors."""


class InvalidAudioProfileError(InvalidProfileError):
    """Video profile errors."""


class InvalidFilterProfileError(InvalidProfileError):
    """Filter profile errors."""


class InvalidOptsError(Exception):
    """Invalid program opts."""


class InvalidChromaDecoderError(Exception):
    """Chroma decoder errors."""


class PipeError(Exception):
    """General pipe errors."""


class PipeNTError(Exception):
    """NT-related pipe errors."""


class ProcessError(Exception):
    """General process errors."""
